from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import torch
from irtorch.rescale import Scale
from irtorch._internal_utils import entropy
from irtorch.outlier_detector import OutlierDetector
from irtorch._internal_utils import random_guessing_data

if TYPE_CHECKING:
    from irtorch.models.base_irt_model import BaseIRTModel

logger = logging.getLogger("irtorch")

class Bit(Scale):
    r"""
    Bit scale transformation, as introduced by :cite:t:`Wallmark2024`.

    Parameters
    ----------
    model : BaseIRTModel
        The IRT model to use for bit scale computation.
    population_theta : torch.Tensor, optional
        Theta scores from the population. Usually the training data.
        Used to find good starting values for the grid of theta scores, which are then used for the bit transformation.
        Recommended to use for models with theta distributions for which values far from 0 are common. (default is None)
    start_theta : torch.Tensor, optional
        The starting theta scores for the bit scale computation. If None, the minimum theta scores are used. (default is None)
    items : list[int], optional
        The item indices for the items to use to compute the bit scores. (default is None and uses all items)
    grid_points : int, optional
        The number of points to use for computing bit score. More steps lead to more accurate results. (default is 400)
    mc_start_theta_approx : bool, optional
        For multiple choice models. Whether to approximate the starting theta scores using simulated random guesses.
        If True, runs :meth:`bit_score_starting_theta_mc`. (default is False)
    **kwargs
        Additional keyword arguments for the starting theta approximation method. See :meth:`bit_score_starting_theta_mc`.
    
    Notes
    -----
    First, item bit scores for each item :math:`j` are computed from :math:`\mathbf{\theta}` scores as follows:

    .. math ::

        \begin{equation}
            \begin{aligned}
                B_j(\mathbf{\theta})=
                \int_{t=\mathbf{\theta}^{(0)}}^{\mathbf{\theta}}
                \left|\frac{dH_j(t)}{dt}\right| dt.
            \end{aligned}
        \end{equation}

    where

    - :math:`\mathbf{\theta}^{(0)}` is the minimum :math:`\mathbf{\theta}`
    - :math:`H_j(\mathbf{\theta})` is entropy for item :math:`j` as a function of :math:`\mathbf{\theta}`
        
    The total bit scores :math:`B(\mathbf{\theta})` are then the sum of the item scores:

    .. math ::

        \begin{equation}
            \begin{aligned}
                B(\mathbf{\theta}) = \sum_{j=1}^{J} B_j(\mathbf{\theta}).
            \end{aligned}
        \end{equation}

    Examples
    --------
    >>> import irtorch
    >>> from irtorch.models import MonotoneNN
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.rescale import Bit
    >>> data, mc_correct = irtorch.load_dataset.swedish_sat_quantitative()
    >>> model = MonotoneNN(data, mc_correct=mc_correct)
    >>> model.fit(train_data=data, algorithm=AE())
    >>> thetas = model.latent_scores(data, theta_estimation="NN")
    >>> # Initalize the scale transformation
    >>> # mc_start_theta_approx sets the starting theta to the approximate score of a randomly guessing respondent
    >>> bit = Bit(model, population_theta=thetas, mc_start_theta_approx=True)
    >>> # Supply the new scale to the model
    >>> model.add_scale_tranformation(bit)
    >>> # Estimate thetas on the transformed scale
    >>> rescaled_thetas = model.latent_scores(data)
    >>> # Or alternatively by directly converting the old ones
    >>> rescaled_thetas = model.transform_theta(thetas)
    >>> # Plot the differences
    >>> model.plot.plot_latent_score_distribution(thetas).show()
    >>> model.plot.plot_latent_score_distribution(rescaled_thetas).show()
    >>> # Plot an item on the bit transformed scale
    >>> model.plot.plot_item_probabilities(1).show()
    """
    def __init__(
        self,
        model: BaseIRTModel,
        population_theta: torch.Tensor = None,
        start_theta: torch.Tensor = None,
        items: list[int] = None,
        grid_points: int = 400,
        mc_start_theta_approx: bool = False,
        **kwargs
    ):
        super().__init__(invertible=False)
        if grid_points <= 0:
            raise ValueError("steps must be a positive integer")
        if items is None:
            items = list(range(len(model.item_categories)))
        elif not isinstance(items, list) or not all(isinstance(item, int) for item in items):
            raise ValueError("items must be a list of integers.")
        
        self.grid_points = grid_points
        self.items = items
        self.model = model
        self._invert_scale_multiplier = self._get_inverted_scale_multiplier()
        self._population_theta = population_theta
        if start_theta is not None:
            if start_theta.flatten().shape != (self.model.latent_variables,):
                raise ValueError("start_theta must have the same number of elements as the number of latent variables.")
            self._start_theta = start_theta.flatten()
        else:
            if mc_start_theta_approx:
                if self.model.mc_correct is None:
                    raise ValueError("Random sampling start theta approximation is only supported for multiple choice models.")
                self.bit_score_starting_theta_mc(**kwargs)
            else:
                self._start_theta = self._invert_scale_multiplier * torch.full((self.model.latent_variables, ), -10000)

    def set_start_theta(self, start_theta: torch.Tensor):
        """
        Sets the starting theta scores for the bit scale computation.

        Parameters
        ----------
        start_theta : torch.Tensor
            The starting theta scores for the bit scale computation.
        """
        if start_theta.shape != (self.model.latent_variables,):
            raise ValueError("start_theta must have the same number of elements as the number of latent variables.")
        self._start_theta = start_theta

    @torch.no_grad()
    def transform(
        self,
        theta: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        Transforms :math:`\mathbf{\theta}` scores into bit scores :math:`B(\mathbf{\theta})`.
        
        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor. Columns are latent variables and rows are respondents.
        
        Returns
        -------
        torch.Tensor
            A 2D tensor with bit score scale scores for each respondent across the rows together with another tensor with start_theta.
        """
        start_theta_adjusted = self._start_theta * self._invert_scale_multiplier
        theta_adjusted = torch.max(theta * self._invert_scale_multiplier, start_theta_adjusted)
        if self._population_theta is None:
            grid_start = torch.full((1, self.model.latent_variables), -7)
            grid_end = torch.full((1, self.model.latent_variables), 7)
            medians = torch.zeros(1, self.model.latent_variables)
        else:
            population_theta_adjusted = torch.max(self._population_theta * self._invert_scale_multiplier , start_theta_adjusted)
            grid_start, grid_end = self._get_grid_boundaries(population_theta_adjusted, start_theta_adjusted)
            medians, _ = torch.median(population_theta_adjusted, dim=0)
        
        bit_scores = self._compute_multi_dimensional_bit_scores(
            theta_adjusted,
            start_theta_adjusted,
            medians,
            grid_start,
            grid_end,
            self._invert_scale_multiplier,
            self.grid_points
        )
        
        return bit_scores

    @torch.no_grad()
    def transform_to_1D(
        self,
        theta: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        Transforms :math:`\mathbf{\theta}` scores of a multi-dimensional model into one-dimensional bit scores :math:`B(\mathbf{\theta})`.
        Equivalent to :meth:`transform` for one-dimensional models.
        
        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor. Columns are latent variables and rows are respondents.
        
        Returns
        -------
        torch.Tensor
            A 2D tensor with bit score scale scores for each respondent across the rows together with another tensor with start_theta.
        """
        if self.model.latent_variables == 1:
            return self.transform(theta=theta)

        start_theta_adjusted = self._start_theta * self._invert_scale_multiplier 
        theta_adjusted = torch.max(theta * self._invert_scale_multiplier, start_theta_adjusted)
        if self._population_theta is None:
            grid_start = torch.full((1, self.model.latent_variables), -7)
            grid_end = torch.full((1, self.model.latent_variables), 7)
        else:
            population_theta_adjusted = torch.max(self._population_theta * self._invert_scale_multiplier , start_theta_adjusted)
            grid_start, grid_end = self._get_grid_boundaries(population_theta_adjusted, start_theta_adjusted)
        
        bit_scores = self._compute_1d_bit_scores(
            theta_adjusted,
            start_theta_adjusted,
            grid_start,
            grid_end,
            self._invert_scale_multiplier,
            self.grid_points
        )
        
        return bit_scores

    def inverse(self, transformed_theta):
        raise NotImplementedError("Bit scale inverse is not implemented.")

    @torch.inference_mode(False)
    def jacobian(
        self,
        theta: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        Computes the gradients of the bit scores with respect to the input theta scores.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A torch tensor with the gradients for each theta score. Dimensions are (theta rows, latent variables, latent variables) where the last two are the jacobians.
        """
        if self.model.latent_variables == 1:
            if theta.requires_grad:
                theta.requires_grad_(False)

            gradients = torch.zeros(theta.shape[0], len(self.model.item_categories), theta.shape[1])
            theta_scores = theta.clone()
            theta_scores.requires_grad_(True)
            probs = self.model.item_probabilities(theta_scores)
            entropies = entropy(probs)

            for item in range(entropies.shape[1]):
                if theta_scores.grad is not None:
                    theta_scores.grad.zero_()
                entropies[:, item].sum().backward(retain_graph=True)
                for latent_variable in range(theta.shape[1]):
                    gradients[:, item, latent_variable] = theta_scores.grad[:, latent_variable]

            gradients[gradients.isnan()] = 0.
            if self.items is not None:
                gradients = gradients[:, self.items, :]

            # sum over items and add dimension for jacobian
            # multiply by -1 if we are inversely related to the theta scores
            return self._invert_scale_multiplier.flatten() * gradients.abs().sum(dim=1).unsqueeze(dim=-1)
        else:
            raise NotImplementedError("Multidimensional bit scale gradients is not implemented.")
        # if hasattr(self.model.algorithm, "training_theta_scores") and self.model.algorithm.training_theta_scores is not None:
        #     q1_q3 = torch.quantile(self.model.algorithm.training_theta_scores, torch.tensor([0.25, 0.75]), dim=0)
        # else:
        #     q1_q3 = torch.tensor([-0.6745, 0.6745]).unsqueeze(1).expand(-1, self.model.latent_variables)

        # iqr = q1_q3[1] - q1_q3[0]
        # lower_bound = q1_q3[0] - 1.5 * iqr
        # upper_bound = q1_q3[1] + 1.5 * iqr
        # h = (upper_bound - lower_bound) / 1000
        # theta_low = theta-h
        # theta_high = theta+h
        # if independent_theta is None:
        #     gradients = torch.zeros(theta.shape[0],theta.shape[1],theta.shape[1])
        #     for latent_variable in range(theta.shape[1]):
        #         theta_low_var = torch.cat((theta[:, :latent_variable], theta_low[:, latent_variable].view(-1, 1), theta[:, latent_variable+1:]), dim=1)
        #         theta_high_var = torch.cat((theta[:, :latent_variable], theta_high[:, latent_variable].view(-1, 1), theta[:, latent_variable+1:]), dim=1)
        #         bit_scores_low = self.transform(
        #             theta_low_var, one_dimensional=one_dimensional, grid_points=grid_points, items=items
        #         )
        #         bit_scores_high = self.transform(
        #             theta_high_var, one_dimensional=one_dimensional, grid_points=grid_points, items=items
        #         )
        #         gradients[:, latent_variable, :] = (bit_scores_high - bit_scores_low) / (2 * h[latent_variable])
        # else:
        #     theta_low_var = torch.cat((theta[:, :independent_theta-1], theta_low[:, independent_theta-1].view(-1, 1), theta[:, independent_theta:]), dim=1)
        #     theta_high_var = torch.cat((theta[:, :independent_theta-1], theta_high[:, independent_theta-1].view(-1, 1), theta[:, independent_theta:]), dim=1)
        #     bit_scores_low = self(theta_low_var, one_dimensional=one_dimensional, grid_points=grid_points, items=items)
        #     bit_scores_high = self(theta_high_var, one_dimensional=one_dimensional, grid_points=grid_points, items=items)
        #     gradients = (bit_scores_high - bit_scores_low) / (2 * h[independent_theta-1])

        # return gradients

    def bit_score_starting_theta_mc(
        self,
        theta_estimation: str = "ML",
        ml_map_device: str = "cuda" if torch.cuda.is_available() else "cpu",
        lbfgs_learning_rate: float = 0.25,
        items: list[int] = None,
        guessing_probabilities: list[float] = None,
        guessing_iterations: int = 10000,
    ):
        r"""
        For multiple choice models, approximate the starting theta score :math:`\mathbf{\theta}^{(0)}` from which to compute bit scores.
        See notes under :meth:`bit_scores` for the bit score formula.
        
        Parameters
        ----------
        model : BaseIRTModel
            The IRT model to use for computing the starting theta scores.
        theta_estimation : str, optional
            Method used to obtain the theta scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'ML')
        ml_map_device: str, optional
            For ML and MAP. The device to use for computation. Can be 'cpu' or 'cuda'. (default is "cuda" if available else "cpu")
        lbfgs_learning_rate: float, optional
            For ML and MAP. The learning rate to use for the LBFGS optimizer. (default is 0.3)
        items: list[int], optional
            The item indices for the items to use to compute the bit scores. (default is None and uses all items)
        guessing_probabilities: list[float], optional
            The guessing probability for each item. The same length as the number of items. Guessing is not supported for polytomously scored items and the probabilities for them will be ignored. (default is None and uses no guessing or, for multiple choice models, 1 over the number of item categories)
        guessing_iterations: int, optional
            The number of iterations to use for approximating a minimum theta when guessing is incorporated. (default is 200)
        
        Returns
        -------
        torch.Tensor
            A tensor with all the starting theta values.
        """
        if self.model.mc_correct is None:
            raise ValueError("This method is only supported for multiple choice models.")

        items = items or list(range(len(self.model.item_categories)))
        mc_correct = torch.tensor(self.model.mc_correct)
        if not all(isinstance(item, int) for item in items):
            raise ValueError("items must be a list of integers.")
        
        selected_item_categories = [self.model.item_categories[i] for i in items]

        if guessing_probabilities:
            if len(guessing_probabilities) != len(items) or not all(0 <= num < 1 for num in guessing_probabilities):
                raise ValueError("guessing_probabilities must be list of floats with the same length as the number of items and values between 0 and 1.")
        else:
            guessing_probabilities = [1 / categories for categories in selected_item_categories]

        selected_correct = mc_correct.index_select(0, torch.tensor(items))
        random_data = random_guessing_data(
            selected_item_categories,
            guessing_iterations,
            guessing_probabilities,
            selected_correct
        )

        logger.info("Approximating theta from random guessing data to get minimum bit score.")
        guessing_theta = self.model.latent_scores(random_data, theta_estimation=theta_estimation, device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate, rescale=False)

        starting_theta = guessing_theta.detach().median(dim=0).values.reshape(1, self.model.latent_variables)

        self._start_theta = starting_theta.flatten()

    def _compute_1d_bit_scores(self, theta_adjusted, start_theta_adjusted, grid_start, grid_end, invert_scale_multiplier, grid_points):
        """
        Computes the 1D bit scores for multidimensional models for the given input theta scores.

        Parameters
        ----------
        theta_adjusted : torch.Tensor
            The adjusted theta-scores. A 2D tensor.
        start_theta_adjusted : torch.Tensor
            The adjusted theta-scores for the starting point. A one row tensor.
        grid_start : torch.Tensor
            The start of the grid. A one row tensor.
        grid_end : torch.Tensor
            The end of the grid. A one row tensor.
        invert_scale_multiplier : torch.Tensor
            A one row tensor with -1 for latent variables inversely related to the test scores and 1 for others.
        grid_points : int
            The number of grid points.

        Returns
        -------
        torch.Tensor
            The computed 1D bit scores.
        """
        # Limit all theta_scores to be within the grid range
        theta_adjusted_capped = torch.clamp(theta_adjusted, grid_start, grid_end)
        
        grid_linespace = torch.linspace(0, 1, steps=grid_points).unsqueeze(-1)
        # Use broadcasting to compute the grid
        grid = grid_start + (grid_linespace * (theta_adjusted_capped - grid_start).unsqueeze(1))

        # set the entire grid for those below the grid to their theta scores
        below_grid_start_mask = theta_adjusted < grid_start
        grid = torch.where(below_grid_start_mask.unsqueeze(1), theta_adjusted.unsqueeze(1), grid)

        # Ensure all values are larger than start_theta_adjusted
        grid = torch.maximum(grid, start_theta_adjusted)

        # set the last slot in each grid to the outliers theta scores
        grid[:, -1, :] = torch.where(theta_adjusted > grid_end, theta_adjusted, grid[:, -1])
        # set the first slot in each grid to the start_theta_adjusted (required when theta_adjusted is an outlier)
        grid[:, 0, :] = start_theta_adjusted.squeeze()

        # convert back to non-inverted theta scale and compute grid entropies
        grid = grid.view(-1, grid.shape[2]) * invert_scale_multiplier
        output = self.model(grid)
        entropies = entropy(self.model.probabilities_from_output(output))[:, self.items]

        # the goal is to get the entropies to the dimensions required for bit_score_distance
        # we need to transpose for view to order them correctly
        # and change positioning of each dimension in the end using permute
        entropies = (
            entropies.t()
            .view(entropies.shape[1], int(entropies.shape[0] / grid_points), grid_points)
            .permute(1, 0, 2)
        )

        # Compute the absolute difference
        diff = torch.abs(entropies - torch.roll(entropies, shifts=1, dims=2))
        # Note that for each sub-tensor in the third dimension, the first element will
        # be the difference with the last element of the previous sub-tensor because
        # of the roll function. We set this to 0
        diff[:, :, 0] = 0
        return diff.sum(dim=(1, 2)).unsqueeze(1)

    def _compute_multi_dimensional_bit_scores(self, theta_adjusted, start_theta_adjusted, median_thetas, grid_start, grid_end, invert_scale_multiplier, grid_points):
        """
        Computes the multi-dimensional bit scores for the given input theta scores (or more efficient one-dimensional scores for 1D models).

        Parameters
        -----------
        theta_adjusted : torch.Tensor
            The input theta scores.
        start_theta_adjusted : torch.Tensor
            The minimum theta score to be used in the grid. A one row tensor.
        median_thetas : torch.Tensor
            The median theta score of each latent variable.
        grid_start : float
            The minimum value of the grid. A one row tensor.
        grid_end : float
            The maximum value of the grid. A one row tensor.
        invert_scale_multiplier : torch.Tensor
            A one row tensor with -1 for latent variables inversely related to the test scores and 1 for others.
        grid_points : int
            The number of points in the grid.

        Returns
        --------
        bit_scores : torch.Tensor
            The multi-dimensional bit scores for the input theta scores.
        """
        # Construct grid
        ratio = torch.linspace(0, 1, steps=grid_points).view(-1, 1)
        grid = (1 - ratio) * grid_start + ratio * grid_end
        # Add theta scores to the grid and sort the columns
        grid = torch.cat([grid, theta_adjusted], dim=0)
        # set all theta scores smaller than start_theta_adjusted to start_theta_adjusted
        grid = torch.max(grid, start_theta_adjusted)
        # set the first slot in each grid to the start_theta_adjusted
        # required when any value in theta_adjusted is an outlier
        grid[0, :] = start_theta_adjusted

        grid, sorted_indices = torch.sort(grid, dim=0)
        # Fill each column in grid we are not not computing bit scores for with the median
        bit_scores = torch.zeros_like(theta_adjusted)
        for theta_var in range(grid.shape[1]):
            # Only compute once per unique value
            unique_grid, inverse_indices = grid[:, theta_var].unique(return_inverse=True)
            latent_variable_grid = median_thetas.repeat(unique_grid.shape[0], 1)
            latent_variable_grid[:, theta_var] = unique_grid

            # Convert back to non-inverted theta scale and compute grid entropies
            output = self.model(latent_variable_grid * invert_scale_multiplier)
            entropies = entropy(self.model.probabilities_from_output(output))[:, self.items]

            # Compute the absolute difference between each grid point entropy and the previous one
            diff = torch.zeros_like(entropies)
            diff[1:,:] = torch.abs(entropies[:-1, :] - entropies[1:, :])

            # cummulative sum over grid points and then sum the item scores
            bit_score_grid = diff.sum(dim=1).cumsum(dim=0)

            # add duplicates, unsort and take only the bit scores for the input theta scores
            bit_scores[:, theta_var] = bit_score_grid[inverse_indices][torch.sort(sorted_indices[:, theta_var])[1]][-theta_adjusted.shape[0]:]
            
        return bit_scores


    def _get_inverted_scale_multiplier(self):
        """
        Compute a tensor with information about whether each latent variable is positively or negatively related to the test scores.
        Assumes monotonicity of the IRT model in the theta range (-1, 1).

        Returns
        -------
        torch.Tensor
            A one row tensor with elements corresponding to latent variables. 1's for positive and -1's for negative relationships.

        Notes
        -----
        If the neural network is set to handle multiple choice correct answers, 
        the method computes the scores based on the correct responses. 
        Otherwise, it simply sums up the training data. 
        The method then performs a linear regression between the latent variables 
        and the scores, and inverts the scales based on the linear weights.
        """
        inverted_scale_multiplier = torch.ones(1, self.model.latent_variables)
        for latent_variable in range(self.model.latent_variables):
            thetas = torch.zeros(2, self.model.latent_variables)
            thetas[0, latent_variable] = -1.
            thetas[1, latent_variable] = 1.
            scores = self.model.expected_scores(thetas).sum(dim=1)
            if scores[0] > scores[1]:
                inverted_scale_multiplier[0, latent_variable] = -1.

        return inverted_scale_multiplier


    def _get_grid_boundaries(self, train_theta_adjusted: torch.Tensor, start_theta_adjusted):
        """
        Determines the start and end points of the grid used for computing bit scores.

        Parameters
        ----------
        train_theta_adjusted : torch.Tensor
            A 2D array containing the adjusted theta scores of the training data. Each row represents one respondent, and each column represents a latent variable.
        start_theta_adjusted : torch.Tensor
            A 1 row tensor containing starting theta scores for the bit scores.

        Returns
        -------
        grid_start : torch.Tensor
            A 1 row tensor containing the start points of the grid for each latent variable.
        grid_end : torch.Tensor
            A 1 row tensor containing the end points of the grid for each latent variable.
        """        
        outlier_detector = OutlierDetector(factor=4)
        start_is_outlier = outlier_detector.is_outlier(start_theta_adjusted, data=train_theta_adjusted, lower=True)[0, :]
        if any(start_is_outlier):
            smallest_non_outlier = outlier_detector.smallest_largest_non_outlier(train_theta_adjusted, smallest=True)
            grid_start = torch.max(start_theta_adjusted, smallest_non_outlier)
        else:
            grid_start = start_theta_adjusted
        grid_end = outlier_detector.smallest_largest_non_outlier(train_theta_adjusted, smallest=False)
        return grid_start, grid_end


    # def bit_score_starting_theta(
    #     self,
    #     theta_estimation: str = "ML",
    #     ml_map_device: str = "cuda" if torch.cuda.is_available() else "cpu",
    #     lbfgs_learning_rate: float = 0.25,
    #     items: list[int] = None,
    #     start_all_incorrect: bool = False,
    #     train_theta: torch.Tensor = None,
    #     guessing_probabilities: list[float] = None,
    #     guessing_iterations: int = 10000,
    # ):
    #     r"""
    #     Computes the starting theta score :math:`\mathbf{\theta}^{(0)}` from which to compute bit scores. See notes under :meth:`bit_scores` for more details.
        
    #     Parameters
    #     ----------
    #     train_theta : torch.Tensor
    #         A 2D tensor with the training data theta scores. Used to estimate relationships between theta and getting the items correct when start_all_incorrect is False. Columns are latent variables and rows are respondents.
    #     theta_estimation : str, optional
    #         Method used to obtain the theta scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'ML')
    #     ml_map_device: str, optional
    #         For ML and MAP. The device to use for computation. Can be 'cpu' or 'cuda'. (default is "cuda" if available else "cpu")
    #     lbfgs_learning_rate: float, optional
    #         For ML and MAP. The learning rate to use for the LBFGS optimizer. (default is 0.3)
    #     items: list[int], optional
    #         The item indices for the items to use to compute the bit scores. (default is None and uses all items)
    #     start_all_incorrect: bool, optional
    #         Whether to compute the starting theta scores based incorrect responses. If false, starting theta is computed based on relationships between each latent variable and the item responses. (default is False)
    #     guessing_probabilities: list[float], optional
    #         The guessing probability for each item. The same length as the number of items. Guessing is not supported for polytomously scored items and the probabilities for them will be ignored. (default is None and uses no guessing or, for multiple choice models, 1 over the number of item categories)
    #     guessing_iterations: int, optional
    #         The number of iterations to use for approximating a minimum theta when guessing is incorporated. (default is 200)
        
    #     Returns
    #     -------
    #     torch.Tensor
    #         A tensor with all the starting theta values.
    #     """
    #     items = items or list(range(len(self.model.item_categories)))
    #     mc_correct = torch.tensor(self.model.mc_correct) if self.model.mc_correct else None
    #     if not all(isinstance(item, int) for item in items):
    #         raise ValueError("items must be a list of integers.")
        
    #     if guessing_probabilities:
    #         if len(guessing_probabilities) != len(items) or not all(0 <= num < 1 for num in guessing_probabilities):
    #             raise ValueError("guessing_probabilities must be list of floats with the same length as the number of items and values between 0 and 1.")

    #     selected_item_categories = [self.model.item_categories[i] for i in items]

    #     if guessing_probabilities is None and mc_correct is not None:
    #         guessing_probabilities = [1 / categories for categories in selected_item_categories]

    #     if not start_all_incorrect:
    #         # Which latent variables are inversely related to the test scores?
    #         item_sum_scores = self.model.expected_scores(train_theta, return_item_scores=False)
    #         test_weights = linear_regression(train_theta, item_sum_scores.reshape(-1, 1))[1:]
    #         inverted_scale = torch.where(test_weights < 0, torch.tensor(-1), torch.tensor(1)).reshape(-1)
            
    #         # Which latent variables are positively related to the item scores?
    #         directions = self.model.item_theta_relationship_directions(train_theta)
    #         item_theta_positive = (inverted_scale * directions) >= 0 # Invert item relationship if overall test relationship is inverted

    #     if guessing_probabilities is None:
    #         if start_all_incorrect:
    #             logger.info("Computing theta for all incorrect responses to get minimum bit score.")
    #             starting_theta = self.model.latent_scores(torch.zeros((1, len(items))).float(), theta_estimation=theta_estimation, ml_map_device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate, rescale=False)
    #         else:
    #             # Get minimum score in relation to each latent variable
    #             min_sum_score = torch.zeros((len(items), self.model.latent_variables))
    #             min_sum_score[~item_theta_positive] = (torch.tensor(selected_item_categories) - 1).view(-1, 1).float().repeat(1, self.model.latent_variables)[~item_theta_positive]

    #             # get the minimum theta scores based on the sum scores
    #             starting_theta = torch.zeros((1, self.model.latent_variables)).float()
    #             for theta in range(self.model.latent_variables):
    #                 logger.info("Computing minimum bit score theta for latent variable %s.", theta+1)
    #                 starting_theta[:, theta] = self.model.latent_scores(min_sum_score.float()[:, theta], theta_estimation=theta_estimation, ml_map_device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate, rescale=False)[:, theta]
    #     else:
    #         if mc_correct is None:
    #             selected_correct = torch.ones(len(items))
    #             random_data = random_guessing_data(selected_item_categories, guessing_iterations, guessing_probabilities)
    #         else:
    #             selected_correct = mc_correct.index_select(0, torch.tensor(items))
    #             random_data = random_guessing_data(
    #                 selected_item_categories,
    #                 guessing_iterations,
    #                 guessing_probabilities,
    #                 selected_correct
    #             )

    #         if start_all_incorrect:
    #             # With one-dimensional bit scores, guessing for all items makes sense
    #             logger.info("Approximating theta from random guessing data to get minimum bit score.")
    #             guessing_theta = self.model.latent_scores(random_data, theta_estimation=theta_estimation, ml_map_device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate, rescale=False)
    #         else:
    #             # Random for positively related and set to correct for others
    #             guessing_theta = torch.zeros(random_data.shape[0], self.model.latent_variables)
    #             for theta in range(self.model.latent_variables):
    #                 random_data_theta = random_data.clone()
    #                 random_data_theta[:, ~item_theta_positive[:, theta]] = selected_correct[~item_theta_positive[:, theta]].float() - 1
    #                 logger.info("Approximating minimum bit score theta from random guessing data for latent variable %s.", theta+1)
    #                 guessing_theta[:, theta] = self.model.latent_scores(random_data_theta, theta_estimation=theta_estimation, ml_map_device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate, rescale=False)[:, theta]

    #         starting_theta = guessing_theta.detach().median(dim=0).values.reshape(1, self.model.latent_variables)

    #     return starting_theta


import pandas as pd
import torch
import torch.nn as nn
from irtorch.models.base_irt_model import BaseIRTModel

class ThreeParameterLogistic(BaseIRTModel):
    r"""
    Three parametric logistic (3PL) IRT model :cite:p:`Birnbaum1968`. *Note that parameter estimation for this model is currently very unstable.*

    Parameters
    ----------
    data: torch.Tensor, optional
        A 2D torch tensor with test data. Used to automatically compute the number of items. Columns are items and rows are respondents. (default is None)
    latent_variables : int
        Number of latent variables.
    items : int
        Number of items.
    item_theta_relationships : torch.Tensor, optional
        A boolean tensor of shape (items, latent_variables). If specified, the model will have connections between latent dimensions and items where the tensor is True. If left out, all latent variables and items are related (Default: None)
    
    Notes
    -----
    For an item :math:`j`, the model defines the probability for responding correctly as:

    .. math::

        c_j+(1-c_j)
        \frac{
            \exp(\mathbf{a}_j^\top \mathbf{\theta} + d_j)
        }{
            1+\exp(\mathbf{a}_j^\top \mathbf{\theta} + d_j)
        },

    where:

    - :math:`\mathbf{\theta}` is a vector of latent variables.
    - :math:`\mathbf{a}_j` is a vector of weights for item :math:`j`.
    - :math:`d_j` is the bias term for item :math:`j`.
    - :math:`c_j` is the probably of guessing correctly on item :math:`j`.

    Examples
    --------
    >>> from irtorch.models import ThreeParameterLogistic
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.load_dataset import swedish_sat_binary
    >>> # Use quantitative part of the SAT data
    >>> data = swedish_sat_binary()[:, :80]
    >>> model = ThreeParameterLogistic(items=80)
    >>> model.fit(train_data=data, algorithm=AE())
    """
    def __init__(
        self,
        data: torch.Tensor = None,
        latent_variables: int = 1,
        items: int = None,
        item_theta_relationships: torch.Tensor = None
    ):
        if items is None and data is None:
            raise ValueError("Either items or data must be provided to initialize the model.")
        if data is not None:
            items = data.size(1)

        super().__init__(latent_variables=latent_variables, item_categories = [2] * items)
        if item_theta_relationships is not None:
            if item_theta_relationships.shape != (items, latent_variables):
                raise ValueError(
                    f"latent_item_connections must have shape ({items}, {latent_variables})."
                )
            assert(item_theta_relationships.dtype == torch.bool), "latent_item_connections must be boolean type."
            assert(torch.all(item_theta_relationships.sum(dim=1) > 0)), "all items must have a relationship with a least one latent variable."
        else:
            item_theta_relationships = torch.ones(items, latent_variables, dtype=torch.bool)

        self.output_size = self.items * 2
        self.guessing_param = nn.Parameter(torch.full((self.items,), -1.1))
        self.weight_param = nn.Parameter(torch.zeros(item_theta_relationships.sum().int()))
        self.bias_param = nn.Parameter(torch.zeros(self.items))

        self.register_buffer("free_weights", item_theta_relationships.bool())
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.ones_(self.weight_param)
        nn.init.zeros_(self.bias_param)
        nn.init.constant_(self.guessing_param, -1.1)

    def forward(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model.

        Parameters
        ----------
        theta : torch.Tensor
            2D tensor with latent variables. Rows are respondents and latent variables are columns. 

        Returns
        -------
        output : torch.Tensor
            2D tensor. Rows are respondents and columns are probabilities.
        """
        bias = self.bias_param

        weights = torch.zeros(self.items, self.latent_variables, device=theta.device)
        weights[self.free_weights] = self.weight_param
        weighted_theta = torch.matmul(theta, weights.T)
        probs = (weighted_theta + bias).sigmoid()
        # Apply sigmoid to guessing_param to keep it in range (0, 1)
        guessing_param = self.guessing_param.sigmoid()
        probs = probs * (1 - guessing_param) + guessing_param

        return probs

    def probabilities_from_output(self, output: torch.Tensor) -> torch.Tensor:
        """
        Compute probabilities from the output tensor from the forward method.

        Parameters
        ----------
        output : torch.Tensor
            Output tensor from the forward pass.

        Returns
        -------
        torch.Tensor
            3D tensor with dimensions (respondents, items, item categories)
        """
        # output is the probability of getting the item correct
        probs = torch.stack([1 - output, output], dim=2)
        return probs

    def log_likelihood(
        self,
        data: torch.Tensor,
        output: torch.Tensor,
        missing_mask: torch.Tensor = None,
        loss_reduction: str = "sum",
    ) -> torch.Tensor:
        """
        Compute the log likelihood.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents
        output: torch.Tensor
            A 2D tensor with output. Columns are item response categories and rows are respondents
        missing_mask: torch.Tensor, optional
            A 2D tensor with missing data mask. (default is None)
        loss_reduction: str, optional 
            The reduction argument. (default is 'sum')
        
        Returns
        -------
        torch.Tensor
            The log likelihood.
        """
        probabilities = self.probabilities_from_output(output)
        data = data.long().view(-1)
        reshaped_probabilities = probabilities.reshape(-1, 2)

        if missing_mask is not None:
            missing_mask = missing_mask.view(-1)
            reshaped_probabilities = reshaped_probabilities[~missing_mask]
            respondents = data.size(0)
            data = data[~missing_mask]

        ll = reshaped_probabilities[torch.arange(data.size(0)), data].log()
        if loss_reduction == "sum":
            return ll.sum()
        elif loss_reduction == "none" and missing_mask is not None:
            ll_masked = torch.full((respondents, ), torch.nan, device= ll.device)
            ll_masked[~missing_mask] = ll
            return ll_masked
        else:
            raise ValueError("loss_reduction must be 'sum' or 'none'")

    def item_parameters(self, irt_format = False) -> pd.DataFrame:
        """
        Get the item parameters for a fitted model.

        Parameters
        ----------
        irt_format : bool, optional
            Only for unidimensional models. Whether to return the item parameters in traditional IRT format. Otherwise returns weights and biases. (default is False)

        Returns
        -------
        pd.DataFrame
            A dataframe with the item parameters.
        """
        if irt_format and self.latent_variables > 1:
            raise ValueError("IRT format is only available for unidimensional models.")

        biases = self.bias_param
        biases = biases.reshape(-1, 1)
        weights = torch.zeros(self.items, self.latent_variables)
        weights[self.free_weights] = self.weight_param

        weights_df = pd.DataFrame(weights.detach().numpy())
        weights_df.columns = [f"a{i+1}" for i in range(weights.shape[1])]
        if irt_format:
            biases_df = pd.DataFrame(-(biases/weights).detach().numpy())
            biases_df.columns = ["b"]
        else:
            biases_df = pd.DataFrame(biases.detach().numpy())
            biases_df.columns = ["d"]

        guessing_df = pd.DataFrame(self.guessing_param.sigmoid().detach().numpy(), columns=["c"])
        parameters = pd.concat([weights_df, biases_df, guessing_df], axis=1)

        return parameters

    @torch.no_grad()
    def item_theta_relationship_directions(self, theta:torch.Tensor = None) -> torch.Tensor:
        """
        Get the relationship directions between each item and latent variable for a fitted model.

        Parameters
        ----------
        theta : torch.Tensor, optional
            Not needed for this model. (default is None)
            
        Returns
        -------
        torch.Tensor
            A 2D tensor with the relationships between the items and latent variables. Items are rows and latent variables are columns.
        """
        weights = torch.zeros(self.items, self.latent_variables)
        weights[self.free_weights] = self.weight_param
        return weights.sign().int()



import pytest
from unittest.mock import patch
import torch
from irtorch.estimation_algorithms import AE
from irtorch.models import MonotoneNN, BaseIRTModel, GeneralizedPartialCredit
from irtorch.estimation_algorithms.encoders import StandardEncoder


# The @pytest.fixture decorator is used to create fixture methods.
# This method is called once per test method that uses it, and its return value is passed to the test method as an argument.
# pytest.fixture with params creates two fixtures, one for the CPU and one for the GPU.
# The ids parameter is used to give the tests meaningful names
class TestAE:
    @pytest.fixture()
    def irt_model(self, latent_variables, item_categories):
        model = GeneralizedPartialCredit(
            latent_variables = latent_variables,
            item_categories = item_categories
        )
        return model

    @pytest.fixture()
    def algorithm(self, irt_model: MonotoneNN, data_loaders, latent_variables, item_categories):
        algorithm = AE()
        algorithm.imputation_method = "zero"
        algorithm.one_hot_encoded = False
        # Set DataLoaders from the fixture
        algorithm.data_loader = data_loaders

        algorithm.encoder = StandardEncoder(
            input_dim=len(item_categories),
            latent_variables=latent_variables,
            hidden_dim=[2 * sum(irt_model.item_categories)]
        )

        return algorithm

    def test_theta_scores(self, algorithm: AE, irt_model: BaseIRTModel, test_data):
        output = algorithm.theta_scores(test_data)
        assert output.shape == (120, irt_model.latent_variables)

    def test__train_step(self, algorithm: AE, irt_model: BaseIRTModel):
        algorithm.optimizer = torch.optim.Adam(
            list(algorithm.encoder.parameters()) + list(irt_model.parameters()), lr=0.1, amsgrad=True
        )
        first_loss = float("inf")
        for epoch in range(5):
            loss = algorithm._train_step(irt_model, epoch)
            if first_loss == float("inf"):
                first_loss = loss
        
        assert loss <= first_loss, "Loss should decrease"

    def test_fit(self, algorithm: AE, irt_model: BaseIRTModel, test_data):
        # Mock the inner functions that would be called during training
        with patch.object(
            algorithm, "_train_step", return_value=torch.tensor(0.5)
        ) as mocked_train_step:
            algorithm.fit(
                model=irt_model,
                train_data=test_data[0:100],
                max_epochs=5
            )

            # Check if inner functions are called
            assert mocked_train_step.call_count == 5



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

class RationalQuadraticSpline(nn.Module):
    """
    This module implements a rational quadratic spline, as described in the paper
    "Neural Spline Flows" by :cite:t:`Durkan2019`.

    Parameters
    ----------
    variables : int
        The number of variables (dimensions) to transform.
    num_bins : int, optional
        The number of bins to use for the spline (default is 14).
    lower_input_bound : float, optional
        The left boundary of the transformation interval (default is 0.0).
    upper_input_bound : float, optional
        The right boundary of the transformation interval (default is 1.0).
    lower_output_bound : float, optional
        The bottom boundary of the transformation interval (default is 0.0).
    upper_output_bound : float, optional
        The top boundary of the transformation interval (default is 1.0).
    min_bin_width : float, optional
        The minimum width of each bin (default is 1e-3).
    min_bin_height : float, optional
        The minimum height of each bin (default is 1e-3).
    min_derivative : float, optional
        The minimum derivative value at the knots (default is 1e-3).
    derivative_outside_lower_input_bound : float, optional
        The derivative value outside the lower input bound (default is None).
    derivative_outside_upper_input_bound : float, optional
        The derivative value outside the upper input bound (default is None).
    """
    def __init__(
        self,
        variables: int,
        num_bins=30,
        lower_input_bound=0.0,
        upper_input_bound=1.0,
        lower_output_bound=0.0,
        upper_output_bound=1.0,
        min_bin_width=1e-3,
        min_bin_height=1e-3,
        min_derivative=1e-3,
        derivative_outside_lower_input_bound=None,
        derivative_outside_upper_input_bound=None,
    ):
        super(RationalQuadraticSpline, self).__init__()
        self.num_bins = num_bins
        self.lower_input_bound = lower_input_bound
        self.upper_input_bound = upper_input_bound
        self.lower_output_bound = lower_output_bound
        self.upper_output_bound = upper_output_bound
        self.min_bin_width = min_bin_width
        self.min_bin_height = min_bin_height
        self.min_derivative = min_derivative
        if derivative_outside_lower_input_bound is None:
            self.derivative_outside_lower_input_bound = \
            (upper_output_bound-lower_output_bound) / (upper_input_bound-lower_input_bound)
        else:
            self.derivative_outside_lower_input_bound = derivative_outside_lower_input_bound
        if derivative_outside_upper_input_bound is None:
            self.derivative_outside_upper_input_bound = \
            (upper_output_bound-lower_output_bound) / (upper_input_bound-lower_input_bound)
        else:
            self.derivative_outside_upper_input_bound = derivative_outside_upper_input_bound

        self.unnormalized_widths = nn.Parameter(torch.rand(variables, num_bins))
        self.unnormalized_heights = nn.Parameter(torch.rand(variables, num_bins))
        self.unnormalized_derivatives = nn.Parameter(torch.rand(variables, num_bins - 1))

    def forward(self, inputs, inverse=False):
        """
        Apply the rational quadratic spline transformation.

        Parameters
        ----------
        inputs : torch.Tensor
            The input tensor to transform.
        inverse : bool, optional
            If True, computes the inverse transformation (default is False).

        Returns
        -------
        outputs : torch.Tensor
            The transformed outputs.
        logabsdet : torch.Tensor
            The logarithm of the absolute value of the determinant of the Jacobian.
        """
        return self._unconstrained_rational_quadratic_spline(
            inputs,
            self.unnormalized_widths.unsqueeze(0).expand(inputs.shape[0], *self.unnormalized_widths.shape),
            self.unnormalized_heights.unsqueeze(0).expand(inputs.shape[0], *self.unnormalized_heights.shape),
            self.unnormalized_derivatives.unsqueeze(0).expand(inputs.shape[0], *self.unnormalized_derivatives.shape),
            inverse=inverse,
        )

    def _unconstrained_rational_quadratic_spline(
        self,
        inputs,
        unnormalized_widths,
        unnormalized_heights,
        unnormalized_derivatives,
        inverse=False,
    ):
        """
        Compute the unconstrained rational quadratic spline transformation (linear relationship outside the interval).

        Parameters
        ----------
        inputs : torch.Tensor
            The input tensor to transform.
        unnormalized_widths : torch.Tensor
            The unnormalized widths for the bins.
        unnormalized_heights : torch.Tensor
            The unnormalized heights for the bins.
        unnormalized_derivatives : torch.Tensor
            The unnormalized derivatives at the knots.
        inverse : bool, optional
            If True, computes the inverse transformation (default is False).

        Returns
        -------
        outputs : torch.Tensor
            The transformed outputs.
        logabsdet : torch.Tensor
            The logarithm of the absolute value of the determinant of the Jacobian.
        """
        # Create masks for input intervals
        if inverse:
            below_lower_bound_mask = inputs < self.lower_output_bound
            above_upper_bound_mask = inputs > self.upper_output_bound
        else:
            below_lower_bound_mask = inputs < self.lower_input_bound
            above_upper_bound_mask = inputs > self.upper_input_bound
        inside_interval_mask = ~below_lower_bound_mask & ~above_upper_bound_mask
        outside_interval_mask = ~inside_interval_mask

        outputs = torch.zeros_like(inputs)
        logabsdet = torch.zeros_like(inputs)

        unnormalized_derivatives = F.pad(unnormalized_derivatives, pad=(1, 1))
        constant = np.log(np.exp(1 - self.min_derivative) - 1)
        unnormalized_derivatives[..., 0] = constant
        unnormalized_derivatives[..., -1] = constant

        if inverse:
            outputs = torch.where(
                below_lower_bound_mask,
                self.lower_input_bound + (inputs - self.lower_output_bound) / self.derivative_outside_lower_input_bound,
                outputs
            )
            outputs = torch.where(
                above_upper_bound_mask,
                self.upper_input_bound + (inputs - self.upper_output_bound) / self.derivative_outside_upper_input_bound,
                outputs
            )
        else:
            outputs = torch.where(
                below_lower_bound_mask,
                self.lower_output_bound + (inputs - self.lower_input_bound) * self.derivative_outside_lower_input_bound,
                outputs
            )
            outputs = torch.where(
                above_upper_bound_mask,
                self.upper_output_bound + (inputs - self.upper_input_bound) * self.derivative_outside_upper_input_bound,
                outputs
            )
        logabsdet = torch.where(
            outside_interval_mask,
            torch.zeros_like(logabsdet),
            logabsdet
        )

        if torch.any(inside_interval_mask):
            (
                outputs[inside_interval_mask],
                logabsdet[inside_interval_mask],
            ) = self._rational_quadratic_spline(
                inputs=inputs[inside_interval_mask],
                unnormalized_widths=unnormalized_widths[inside_interval_mask, :],
                unnormalized_heights=unnormalized_heights[inside_interval_mask, :],
                unnormalized_derivatives=unnormalized_derivatives[inside_interval_mask, :],
                inverse=inverse
            )

        return outputs, logabsdet

    def _rational_quadratic_spline(
        self,
        inputs,
        unnormalized_widths,
        unnormalized_heights,
        unnormalized_derivatives,
        inverse=False
    ):
        """
        Compute the rational quadratic spline transformation.

        Parameters
        ----------
        inputs : torch.Tensor
            The input tensor to transform.
        unnormalized_widths : torch.Tensor
            The unnormalized widths for the bins.
        unnormalized_heights : torch.Tensor
            The unnormalized heights for the bins.
        unnormalized_derivatives : torch.Tensor
            The unnormalized derivatives at the knots.
        inverse : bool, optional
            If True, computes the inverse transformation (default is False).

        Returns
        -------
        outputs : torch.Tensor
            The transformed outputs.
        logabsdet : torch.Tensor
            The logarithm of the absolute value of the determinant of the Jacobian.

        Raises
        ------
        ValueError
            If minimal bin width or height is too large for the number of bins.
        """
        if torch.min(inputs) < self.lower_input_bound or torch.max(inputs) > self.upper_input_bound:
            pass

        num_bins = unnormalized_widths.shape[-1]

        if self.min_bin_width * num_bins > 1.0:
            raise ValueError("Minimal bin width too large for the number of bins")
        if self.min_bin_height * num_bins > 1.0:
            raise ValueError("Minimal bin height too large for the number of bins")

        widths = F.softmax(unnormalized_widths, dim=-1)
        widths = self.min_bin_width + (1 - self.min_bin_width * num_bins) * widths
        cumwidths = torch.cumsum(widths, dim=-1)
        cumwidths = F.pad(cumwidths, pad=(1, 0), mode="constant", value=0.0)
        cumwidths = (self.upper_input_bound - self.lower_input_bound) * cumwidths + self.lower_input_bound
        cumwidths[..., 0] = self.lower_input_bound
        cumwidths[..., -1] = self.upper_input_bound
        widths = cumwidths[..., 1:] - cumwidths[..., :-1]

        derivatives = self.min_derivative + F.softplus(unnormalized_derivatives)

        heights = F.softmax(unnormalized_heights, dim=-1)
        heights = self.min_bin_height + (1 - self.min_bin_height * num_bins) * heights
        cumheights = torch.cumsum(heights, dim=-1)
        cumheights = F.pad(cumheights, pad=(1, 0), mode="constant", value=0.0)
        cumheights = (self.upper_output_bound - self.lower_output_bound) * cumheights + self.lower_output_bound
        cumheights[..., 0] = self.lower_output_bound
        cumheights[..., -1] = self.upper_output_bound
        heights = cumheights[..., 1:] - cumheights[..., :-1]

        if inverse:
            cumheights[..., -1] += 1e-6
            bin_idx = (torch.sum(inputs[..., None] >= cumheights, dim=-1) - 1)[..., None]
        else:
            cumwidths[..., -1] += 1e-6
            bin_idx = (torch.sum(inputs[..., None] >= cumwidths, dim=-1) - 1)[..., None]

        input_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
        input_bin_widths = widths.gather(-1, bin_idx)[..., 0]

        input_cumheights = cumheights.gather(-1, bin_idx)[..., 0]
        delta = heights / widths
        input_delta = delta.gather(-1, bin_idx)[..., 0]

        input_derivatives = derivatives.gather(-1, bin_idx)[..., 0]
        input_derivatives_plus_one = derivatives[..., 1:].gather(-1, bin_idx)[..., 0]

        input_heights = heights.gather(-1, bin_idx)[..., 0]

        if inverse:
            a = (inputs - input_cumheights) * (
                input_derivatives + input_derivatives_plus_one - 2 * input_delta
            ) + input_heights * (input_delta - input_derivatives)
            b = input_heights * input_derivatives - (inputs - input_cumheights) * (
                input_derivatives + input_derivatives_plus_one - 2 * input_delta
            )
            c = -input_delta * (inputs - input_cumheights)

            discriminant = b.pow(2) - 4 * a * c
            assert (discriminant >= 0).all()

            root = (2 * c) / (-b - torch.sqrt(discriminant))
            outputs = root * input_bin_widths + input_cumwidths

            theta_one_minus_theta = root * (1 - root)
            denominator = input_delta + (
                (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
                * theta_one_minus_theta
            )
            derivative_numerator = input_delta.pow(2) * (
                input_derivatives_plus_one * root.pow(2)
                + 2 * input_delta * theta_one_minus_theta
                + input_derivatives * (1 - root).pow(2)
            )
            logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

            return outputs, -logabsdet
        else:
            theta = (inputs - input_cumwidths) / input_bin_widths
            theta_one_minus_theta = theta * (1 - theta)

            numerator = input_heights * (
                input_delta * theta.pow(2) + input_derivatives * theta_one_minus_theta
            )
            denominator = input_delta + (
                (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
                * theta_one_minus_theta
            )
            outputs = input_cumheights + numerator / denominator

            derivative_numerator = input_delta.pow(2) * (
                input_derivatives_plus_one * theta.pow(2)
                + 2 * input_delta * theta_one_minus_theta
                + input_derivatives * (1 - theta).pow(2)
            )
            logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

            return outputs, logabsdet

    @torch.no_grad()
    def plot_spline(self, num_points=1000, spline_idx=0):
        """
        Plot the rational quadratic spline.

        Parameters
        ----------
        spline_idx : int, optional
            The index of the spline to plot (default is 0).
        num_points : int, optional
            The number of points to use for the plot (default is 1000).
        """
        x = torch.linspace(self.lower_input_bound, self.upper_input_bound, num_points)
        y, _ = self.forward(x.unsqueeze(-1).repeat(1, self.unnormalized_widths.shape[0]))
        y = y[:, spline_idx]
        df = pd.DataFrame({
            "x": x.cpu().detach().numpy() if x.is_cuda else x.detach().numpy(), 
            "y": y.cpu().detach().numpy() if y.is_cuda else y.detach().numpy()
        })
        fig = px.line(df, x="x", y="y")

        widths = F.softmax(self.unnormalized_widths, dim=-1)
        widths = self.min_bin_width + (1 - self.min_bin_width * self.num_bins) * widths
        cumwidths = torch.cumsum(widths, dim=-1)
        cumwidths = F.pad(cumwidths, pad=(1, 0), mode="constant", value=0.0)
        cumwidths = (self.upper_input_bound - self.lower_input_bound) * cumwidths + self.lower_input_bound
        cumwidths[..., 0] = self.lower_input_bound
        cumwidths[..., -1] = self.upper_input_bound

        heights = F.softmax(self.unnormalized_heights, dim=-1)
        heights = self.min_bin_height + (1 - self.min_bin_height * self.num_bins) * heights
        cumheights = torch.cumsum(heights, dim=-1)
        cumheights = F.pad(cumheights, pad=(1, 0), mode="constant", value=0.0)
        cumheights = (self.upper_output_bound - self.lower_output_bound) * cumheights + self.lower_output_bound
        cumheights[..., 0] = self.lower_output_bound
        cumheights[..., -1] = self.upper_output_bound

        cumwidths = cumwidths[spline_idx].cpu().flatten().detach().numpy() if cumwidths.is_cuda else cumwidths[spline_idx].flatten().detach().numpy()
        cumheights = cumheights[spline_idx].cpu().flatten().detach().numpy() if cumheights.is_cuda else cumheights[spline_idx].flatten().detach().numpy()

        fig.add_trace(go.Scatter(
            x=cumwidths,
            y=cumheights,
            mode='markers',
            name='Points'
        ))

        return fig



import pytest
import torch
from irtorch.models import BaseIRTModel
from irtorch.estimation_algorithms import JML

class DummyIRTModel(BaseIRTModel):
    def __init__(self, latent_variables=1):
        super().__init__(latent_variables, [2] * 10)

    def forward(self, theta):
        return theta

    def log_likelihood(self, responses, model_out, missing_mask=None, loss_reduction="sum"):
        return torch.sum(model_out)

@pytest.fixture
def setup_jml():
    model = DummyIRTModel()
    train_data = torch.randint(0, 2, (100, 10)).float()
    jml = JML()
    return jml, model, train_data

def test_fit(setup_jml):
    jml, model, train_data = setup_jml
    jml.fit(model, train_data, max_epochs=10, batch_size=10)
    assert len(jml.training_history["train_loss"]) > 0

def test_fit_with_start_thetas(setup_jml):
    jml, model, train_data = setup_jml
    start_thetas = torch.randn(100, 1)
    jml.fit(model, train_data, start_thetas=start_thetas, max_epochs=10, batch_size=10)
    assert len(jml.training_history["train_loss"]) > 0

def test_fit_invalid_start_thetas(setup_jml):
    jml, model, train_data = setup_jml
    start_thetas = torch.randn(50, 1)
    with pytest.raises(ValueError):
        jml.fit(model, train_data, start_thetas=start_thetas, max_epochs=10, batch_size=10)

def test_fit_invalid_latent_variables(setup_jml):
    jml, model, train_data = setup_jml
    start_thetas = torch.randn(100, 2)
    with pytest.raises(ValueError):
        jml.fit(model, train_data, start_thetas=start_thetas, max_epochs=10, batch_size=10)



import torch
import torch.nn as nn
import torch.nn.functional as F
from irtorch.models.base_irt_model import BaseIRTModel

class ModelMix(BaseIRTModel):
    r"""
    A mix of models tied together by a single set of latent variables. 
    This allows for a mix of different IRT models to be used when items have different types or some items don't fit a single model.
    Note that the items in the test data must be ordered in the same way as the models are ordered in the list.

    Parameters
    ----------
    models : list[BaseIRTModel]
        A list of IRT models to be mixed together.

    Examples
    --------
    Example of fitting a mix of Generalized Partial Credit (GPC) and Monotone Neural Network (MNN) models to the Swedish National Mathematics 2 dataset.
    The GPC model is used for the first 14 items and MNN for the last 14 items.
    
    >>> from irtorch.models import GeneralizedPartialCredit, MonotoneNN, ModelMix
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.load_dataset import swedish_national_mathematics_2
    >>> data_math = swedish_national_mathematics_2()
    >>> gpc = GeneralizedPartialCredit(data=data_math[:, :14])
    >>> mnn = MonotoneNN(data = data_math[:, 14:])
    >>> mix_model = ModelMix([gpc, mnn])
    >>> mix_model.fit(train_data=data_math, algorithm=AE())
    """
    def __init__(
        self,
        models: list[BaseIRTModel],
    ):
        item_categories = [item for model in models for item in model.item_categories]
        latent_variables = [model.latent_variables for model in models]
        if not all([latent_variables[0] == latent_variables[i] for i in range(1, len(latent_variables))]):
            raise ValueError("All models must have the same number of latent variables.")

        super().__init__(
            latent_variables=latent_variables[0],
            item_categories = item_categories
        )
        # register each model as a submodule
        self.models = nn.ModuleList(models)

    def reset_parameters(self) -> None:
        nn.init.normal_(self.weight_param, mean=1., std=0.01)
        nn.init.zeros_(self.bias_param)
    
    def forward(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model.

        Parameters
        ----------
        theta : torch.Tensor
            2D tensor with latent variables. Rows are respondents and latent variables are columns. 

        Returns
        -------
        output : torch.Tensor
            2D tensor. Rows are respondents and columns are item category logits.
        """
        outputs = [model(theta) for model in self.models]
        concatenated_output = torch.cat(outputs, dim=1)
        return concatenated_output

    def probabilities_from_output(self, output: torch.Tensor) -> torch.Tensor:
        """
        Compute probabilities from the output tensor from the forward method.

        Parameters
        ----------
        output : torch.Tensor
            Output tensor from the forward pass.

        Returns
        -------
        torch.Tensor
            3D tensor with dimensions (respondents, items, item categories)
        """
        outputs = torch.split(
            output,
            [model.max_item_responses * model.items for model in self.models],
            dim=1
        )
        probabilities = [model.probabilities_from_output(output_part) for model, output_part in zip(self.models, outputs)]
        return self._pad_and_concat(probabilities, dim=1, pad_dim=2)

    def _pad_and_concat(self, tensors, dim, pad_dim):
        # Calculate the maximum size along the pad dimension
        max_size = max(tensor.size(pad_dim) for tensor in tensors)
        
        # Create a list to store padded tensors
        padded_tensors = []
        
        # Iterate over the tensors and pad them to the maximum size
        for tensor in tensors:
            pad_size = max_size - tensor.size(pad_dim)
            if pad_size > 0:
                pad = [0] * (2 * tensor.dim())
                pad[-(2 * pad_dim + 1)] = pad_size
                tensor = F.pad(tensor, pad)
            padded_tensors.append(tensor)
        
        # Concatenate the padded tensors along the specified dimension
        return torch.cat(padded_tensors, dim=dim)

    def item_probabilities(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Returns the probabilities for each possible response for all items.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor. Rows are respondents and latent variables are columns.

        Returns
        -------
        torch.Tensor
            3D tensor with dimensions (respondents, items, item categories)
        """
        output = self(theta)
        return self.probabilities_from_output(output)

    def log_likelihood(
        self,
        data: torch.Tensor,
        output: torch.Tensor,
        missing_mask: torch.Tensor = None,
        loss_reduction: str = "sum",
    ) -> torch.Tensor:
        """
        Compute the log likelihood of the data given the model. This is equivalent to the negative cross entropy loss.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents
        output: torch.Tensor
            A 2D tensor with output. Columns are item response categories and rows are respondents
        missing_mask: torch.Tensor, optional
            A 2D tensor with missing data mask. (default is None)
        loss_reduction: str, optional 
            For most models, this is the reduction argument for torch.nn.functional.cross_entropy(). (default is 'sum')
        
        Returns
        -------
        torch.Tensor
            The log likelihood.
        """
        # run model.log_likelihood for each model
        outputs = torch.split(
            output,
            [model.max_item_responses * model.items for model in self.models],
            dim=1
        )
        data_splits = torch.split(
            data,
            [model.items for model in self.models],
            dim=1
        )
        if missing_mask is None:
            mask_splits = [None] * len(self.models)
        else:
            mask_splits = torch.split(
                missing_mask,
                [model.items for model in self.models],
                dim=1
            )
            mask_splits = [mask.contiguous() for mask in mask_splits]
        log_likelihoods = [
            model.log_likelihood(data_split, output, mask_split, loss_reduction)
            for model, data_split, mask_split, output in zip(self.models, data_splits, mask_splits, outputs)
        ]

        if loss_reduction == "mean":
            raise NotImplementedError("loss_reduction='mean' is not implemented for ModelMix.")
        elif loss_reduction == "sum":
            return torch.sum(torch.stack(log_likelihoods), dim=0)
        elif loss_reduction == "none":
            return torch.cat(log_likelihoods, dim=0)



import pandas as pd
import torch
import torch.nn as nn
from irtorch.models.base_irt_model import BaseIRTModel

class NominalResponse(BaseIRTModel):
    r"""
    Nominal response IRT model :cite:p:`Bock1972`.

    Parameters
    ----------
    data: torch.Tensor, optional
        A 2D torch tensor with test data. Used to automatically compute item_categories. Columns are items and rows are respondents. (default is None)
    latent_variables : int, optional
        Number of latent variables. (default is 1)
    item_categories : list[int], optional
        Number of categories for each item. One integer for each item. Missing responses exluded. (default is None)
    item_theta_relationships : torch.Tensor, optional
        A boolean tensor of shape (items, latent_variables). If specified, the model will have connections between latent dimensions and items where the tensor is True. If left out, all latent variables and items are related (Default: None)
    mc_correct : list[int], optional
        Only for multiple choice data. The correct response option for each item. (Default: None)

    Notes
    -----
    For an item :math:`j` with :math:`m=0, 1, 2, \ldots, M_j` possible item responses/scores, the model defines the probability for responding with a score of :math:`x` as follows (selecting response option :math:`x` for multiple choice items):

    .. math::

        P(X_j=x | \mathbf{\theta}) = \frac{
            \exp(\mathbf{a}_{jm}^\top \mathbf{\theta} + d_{jm})
        }{
            \sum_{m=0}^{M_j}
                \exp(\mathbf{a}_{jm}^\top \mathbf{\theta} + d_{jm})
        },

    where:

    - :math:`\mathbf{\theta}` is a vector of latent variables.
    - :math:`\mathbf{a}_{jm}` is a vector of weights for item :math:`j` and response category :math:`m`.
    - :math:`d_{jm}` is the bias term for item :math:`j` and response category :math:`m`.

    The model is fitted with the following constraints of the item parameters for model identifiability:

    .. math::

        \sum_{m=0}^{M_j} a_{jm} = 0, \quad \sum_{m=0}^{M_j} d_{jm} = 0.
    """
    def __init__(
        self,
        data: torch.Tensor = None,
        latent_variables: int = 1,
        item_categories: list[int] = None,
        item_theta_relationships: torch.Tensor = None,
        mc_correct: list[int] = None,
    ):
        if item_categories is None:
            if data is None:
                raise ValueError("Either item_categories or data must be provided to initialize the model.")
            else:
                # replace nan with -inf to get max
                item_categories = (torch.where(~data.isnan(), data, torch.tensor(float('-inf'))).max(dim=0).values + 1).int().tolist()

        super().__init__(latent_variables=latent_variables, item_categories=item_categories, mc_correct=mc_correct)
        if item_theta_relationships is not None:
            if item_theta_relationships.shape != (len(item_categories), latent_variables):
                raise ValueError(
                    f"latent_item_connections must have shape ({len(item_categories)}, {latent_variables})."
                )
            assert(item_theta_relationships.dtype == torch.bool), "latent_item_connections must be boolean type."
            assert(torch.all(item_theta_relationships.sum(dim=1) > 0)), "all items must have a relationship with a least one latent variable."

        self.output_size = self.items * self.max_item_responses

        free_weights = torch.zeros(self.items, self.max_item_responses, latent_variables)
        for item, item_cat in enumerate(self.item_categories):
            if item_theta_relationships is not None:
                free_weights[item, 1:item_cat, :] = item_theta_relationships[item, :]
            else:
                free_weights[item, 1:item_cat, :] = 1.0

        free_weights = free_weights.reshape(-1, latent_variables)
        self.weight_param = nn.Parameter(torch.zeros(free_weights.sum().int()))

        number_of_bias_parameters = sum(self.item_categories) - self.items
        self.bias_param = nn.Parameter(torch.zeros(number_of_bias_parameters))
        missing_category = torch.zeros(self.items, self.max_item_responses)
        for item, item_cat in enumerate(self.item_categories):
            missing_category[item, item_cat:self.max_item_responses] = 1.0
        missing_category = missing_category.reshape(-1)

        first_category = torch.zeros(self.items, self.max_item_responses)
        first_category[:, 0] = 1.0
        first_category = first_category.reshape(-1)
        free_bias = (1 - first_category) * (1 - missing_category)
        self.register_buffer("free_weights", free_weights.bool())
        self.register_buffer("free_bias", free_bias.bool())
        self.register_buffer("missing_category", missing_category.bool())
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.zeros_(self.weight_param)
        nn.init.zeros_(self.bias_param)

    def forward(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model.

        Parameters
        ----------
        theta : torch.Tensor
            2D tensor with latent variables. Rows are respondents and latent variables are columns. 

        Returns
        -------
        output : torch.Tensor
            2D tensor. Rows are respondents and columns are item category logits.
        """
        bias = torch.zeros(self.output_size, device=theta.device)
        bias[self.free_bias] = self.bias_param
        bias = bias.view(self.items, -1)
        bias[:, 0] = -bias[:, 1:].sum(dim=1) # all biases sum to 0 for identifiability

        weights = torch.zeros(self.output_size, self.latent_variables, device=theta.device)
        weights[self.free_weights] = self.weight_param
        weights = weights.view(self.items, -1, self.latent_variables)
        weights[:, 0] = -weights[:, 1:].sum(dim=1) # all weights sum to 0 for identifiability
        weighted_theta = torch.matmul(theta, weights.view(-1, self.latent_variables).T)

        output = weighted_theta + bias.flatten()
        # stop gradients from flowing through the missing categories
        output[:, self.missing_category] = -torch.inf

        return output

    def item_parameters(self) -> pd.DataFrame:
        """
        Get the item parameters for a fitted model.

        Returns
        -------
        pd.DataFrame
            A dataframe with the item parameters.
        """
        biases = torch.zeros(self.output_size)
        biases[self.free_bias] = self.bias_param
        biases = biases.reshape(-1, self.max_item_responses)
        biases[:, 0] = -biases[:, 1:].sum(dim=1)

        weights = torch.zeros(self.output_size, self.latent_variables)
        weights[self.free_weights] = self.weight_param
        weights = weights.reshape(-1, self.max_item_responses, self.latent_variables)
        weights[:, 0] = -weights[:, 1:].sum(dim=1)
        weights = weights.permute(0, 2, 1).reshape(-1, self.max_item_responses * self.latent_variables)

        weights_df = pd.DataFrame(weights.detach().numpy())
        weights_df.columns = [f"a{i+1}{j+1}" for i in range(self.latent_variables) for j in range(int(weights.shape[1]/self.latent_variables))]
        biases_df = pd.DataFrame(biases.detach().numpy())
            
        biases_df.columns = [f"d{i+1}" for i in range(biases_df.shape[1])]
        parameters = pd.concat([weights_df, biases_df], axis=1)

        return parameters

    @torch.no_grad()
    def item_theta_relationship_directions(self, theta:torch.Tensor = None) -> torch.Tensor:
        """
        Get the relationship directions between each item and latent variable for a fitted model.

        Parameters
        ----------
        theta : torch.Tensor, optional
            Only needed for NR models. A 2D tensor with latent theta scores from the population of interest. Each row represents one respondent, and each column represents a latent variable.

        Returns
        -------
        torch.Tensor
            A 2D tensor with the relationships between the items and latent variables. Items are rows and latent variables are columns.
        """
        if theta is None:
            raise ValueError("theta must be provided to get item to theta relationships for NR models.")
        return super().item_theta_relationship_directions(theta)



import torch
from torch import nn
import torch.nn.functional as F

class SoftplusLinear(nn.Module):
    """
    A linear layer with positive weights ensured using the softplus function. Should be more stable than squaring and maintains gradients in contrast to clipping/abs.
    """
    def __init__(
            self,
            in_features,
            out_features,
            separate_inputs: torch.Tensor=None,
            separate_outputs: torch.Tensor=None,
            zero_inputs: torch.Tensor=None,
            zero_outputs: torch.Tensor=None,
            softplus: bool=True
        ):
        """
        Parameters
        ----------
        in_features : int
            Number of input features.
        out_features : int
            Number of output features.
        separate_inputs : torch.Tensor, optional
            A tensor with integers summing up to in_features that determines if the inputs should be separated into multiple groups. 
            If specified, the layer will have multiple groups of inputs/outputs with separate weights. (Default: None)
        separate_outputs : torch.Tensor, optional
            Only when separate inputs is specified. A tensor with integers summing up to out_features.
            Has to have the same number of groups as separate_inputs.
            If specified, the layer will have multiple groups of inputs/outputs with separate weights. 
            (Default: None)
        zero_inputs : torch.Tensor, optional
            A boolean tensor of shape (in_features,) that determines for which inputs weights should be trained. Mostly relevant when training polytomous IRT models for which each response category has a separate network. (Default: None)
        zero_outputs : torch.Tensor, optional
            A boolean tensor of shape (out_features,) that determines which outputs should be zero. Typically used for IRT models to remove relationships between some items or item cateogires and latent variables. (Default: None)
        softplus : bool, optional
            Whether to use the softplus function or not. If False, the raw weights are used like regular linear layer. (Default: True)
        """
        super().__init__()
        if zero_outputs is None:
            zero_outputs = torch.zeros(out_features).int()
        if zero_inputs is None:
            zero_inputs = torch.zeros(in_features).int()
        if separate_inputs is not None:
            if separate_inputs.sum() != in_features:
                raise ValueError("separate_inputs must sum to in_features")
            if separate_outputs is None:
                # Split out_features into the input groups as evenly as possible
                quotient, remainder = divmod(out_features, separate_inputs.shape[0])
                groups = [quotient + 1] * remainder + [quotient] * (separate_inputs.shape[0] - remainder)
                separate_outputs = torch.tensor(groups)
            if separate_outputs.sum() != out_features:
                raise ValueError("separate_outputs must sum to out_features")
        elif separate_outputs is not None:
            raise ValueError("separate_outputs must be None if separate_inputs is None")
        
        self.in_features = in_features
        self.out_features = out_features
        self.softplus = softplus
        self.zero_inputs = zero_inputs
        self.zero_outputs = zero_outputs
        
        if separate_inputs is not None:
            free_weight = self.separate_weights(in_features, out_features, separate_inputs, separate_outputs)
        else:
            free_weight = torch.full((out_features, in_features), True)

        free_weight[zero_outputs.bool()] = False
        free_weight[:, zero_inputs.bool()] = False
        self.register_buffer("free_bias", (1 - zero_outputs).bool())
        self.register_buffer("free_weight", free_weight)

        # These are learnable parameters representing the raw weights of the layer.
        # The actual weights used in the layer's operations are the softplus of these raw weights.
        self.raw_weight_param = nn.Parameter(torch.empty(self.free_weight.sum()))
        self.bias_param = nn.Parameter(torch.empty(self.free_bias.sum()))

        self.reset_parameters()

    @staticmethod
    def separate_weights(in_features, out_features, separate_inputs, separate_outputs):
        free_weights = torch.full((out_features, in_features), False)

        input_indices = torch.cumsum(separate_inputs, dim=0) - separate_inputs
        output_indices = torch.cumsum(separate_outputs, dim=0) - separate_outputs

        for input_group_size, output_group_size, input_idx, output_idx in zip(separate_inputs, separate_outputs, input_indices, output_indices):
            free_weights[output_idx:output_idx+output_group_size, input_idx:input_idx+input_group_size] = True

        return free_weights
        
    def reset_parameters(self):
        """
        Reset the parameters of the layer.
        """
        nn.init.zeros_(self.raw_weight_param)
        nn.init.zeros_(self.bias_param)

    def forward(self, x):
        """
        Forward pass of the layer.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        output : torch.Tensor
            Output tensor.
        """
        weight = torch.zeros((self.out_features, self.in_features), device=x.device)
        bias = torch.zeros(self.out_features, device=x.device)
        if self.softplus:
            weight[self.free_weight] = F.softplus(self.raw_weight_param)
        else:
            weight[self.free_weight] = self.raw_weight_param
        bias[self.free_bias] = self.bias_param
        output = F.linear(x, weight, bias)
        return output



from unittest.mock import MagicMock, patch
import torch
import numpy as np
import pandas as pd
import pytest
from irtorch.models import BaseIRTModel
from irtorch.evaluator import Evaluator
from irtorch.estimation_algorithms import AE
from irtorch.quantile_mv_normal import QuantileMVNormal
from irtorch.gaussian_mixture_model import GaussianMixtureModel

@pytest.fixture
def algorithm(latent_variables):
    item_categories = [2, 3]
    mock_algorithm = MagicMock(spec=AE)
    mock_algorithm.one_hot_encoded = False
    mock_algorithm.train_data = torch.cat(
        [
            torch.randint(0, item_cat, (30, 1))
            for item_cat in item_categories
        ],
        dim=1,
    )
    mock_algorithm.training_theta_scores = torch.randn(30, latent_variables)
    # Mock theta_scores method based on input
    def theta_scores_mock(input_tensor):
        theta_scores = torch.randn(input_tensor.shape[0], latent_variables)
        return theta_scores

    mock_algorithm.theta_scores = MagicMock(side_effect=theta_scores_mock)
    return mock_algorithm

@pytest.fixture
def irt_model(latent_variables, algorithm: AE):
    item_categories = [2, 3]

    def model_forward_mock(input_tensor: torch.Tensor):
        logits = [
            torch.randn(input_tensor.shape[0], category)
            for category in item_categories
        ]
        return torch.cat(logits, dim=1)

    # Mock item_probabilities method based on input
    def item_probabilities_mock(input_tensor):
        logits = torch.randn(input_tensor.shape[0], len(item_categories), max(item_categories))
        for item_id, item_cats in enumerate(item_categories):
            logits[:, item_id, item_cats:] = -torch.inf
        return torch.softmax(logits, dim = 2)
    
    # Mock theta_scores method based on input
    def latent_scores(data, **kwargs):
        theta_scores = torch.randn(data.shape[0], latent_variables)
        return theta_scores
    
    def transform_theta_mock(theta):
        return torch.randn(theta.shape[0],theta.shape[1]).abs() * 10

    mock_model = MagicMock(spec=BaseIRTModel, side_effect=model_forward_mock)
    mock_model.algorithm = algorithm
    mock_model.scale = [5]
    mock_model.transform_theta = MagicMock(side_effect=transform_theta_mock)
    mock_model.item_probabilities = MagicMock(side_effect=item_probabilities_mock)
    mock_model.latent_scores = MagicMock(side_effect=latent_scores)
    mock_model.item_categories = item_categories
    mock_model.item_categories = item_categories
    mock_model.mc_correct = None
    mock_model.latent_variables = latent_variables

    return mock_model

@pytest.fixture
def evaluation(irt_model: BaseIRTModel):
    def pdf_mock(theta):
        return torch.rand(theta.shape[0])

    evaluation = Evaluator(irt_model)
    evaluation.latent_density = QuantileMVNormal()
    evaluation.latent_density.pdf = MagicMock(side_effect=pdf_mock)

    return evaluation

def test_mutual_information_difference(evaluation: Evaluator):
    data = torch.tensor([[1.0, 0.0], [1.0, 0.0], [0.0, 2.0], [1.0, 2.0]])
    
    def probabilities_from_output_mock(input_tensor):
        # tensor that matches the shape of the data tensor
        return torch.tensor([
            [[0.55, 0.45, 0.0], [0.20, 0.35, 0.45]],
            [[0.35, 0.65, 0.0], [0.10, 0.40, 0.50]],
            [[0.25, 0.75, 0.0], [0.10, 0.20, 0.70]],
            [[0.15, 0.85, 0.0], [0.05, 0.05, 0.90]],
        ])

    evaluation.model.probabilities_from_output = MagicMock(side_effect=probabilities_from_output_mock)
    
    mid, amid, _ = evaluation.mutual_information_difference(data=data)
    assert mid.equals(amid), "Mid and Amid are not equal"
    expected_mid = pd.DataFrame([
        [0.8042, 0.3031],
        [0.3031, 0.9862]
    ])
    assert np.allclose(mid.values, expected_mid.values, atol=1e-4), "Mid is not correct"

def test__evaluate_data_theta_input(evaluation: Evaluator):
    # Create some synthetic test data
    data = torch.cat(
        [
            torch.randint(0, item_cat, (30, 1))
            for item_cat in evaluation.model.item_categories
        ],
        dim=1,
    ).float()
    data[5:7, 1] = torch.nan

    # Call the method with data and theta as None
    result_data, result_theta, missing_mask = evaluation._evaluate_data_theta_input(data, None, theta_estimation="NN")

    # Check the shape of the output data and theta
    assert result_data.shape == data.shape
    assert result_theta.shape == evaluation.model.algorithm.training_theta_scores.shape
    assert missing_mask.shape == data.shape
    assert missing_mask[5:7, 1].all()

    # Call the method with data as None
    result_data, result_theta, missing_mask = evaluation._evaluate_data_theta_input(None, None, theta_estimation="NN")

    # Check the shape of the output data and theta
    assert result_data.shape == evaluation.model.algorithm.train_data.shape
    assert result_theta.shape == evaluation.model.algorithm.training_theta_scores.shape
    assert missing_mask.shape == data.shape
    assert (~missing_mask).all()

def test_probabilities_from_grouped_theta(evaluation: Evaluator, latent_variables):
    # Create some synthetic theta scores
    grouped_theta = (
        torch.randn(10, latent_variables),
        torch.randn(10, latent_variables),
        torch.randn(10, latent_variables),
    )

    # Call the method
    probabilities = evaluation._grouped_theta_probabilities(grouped_theta)
    assert probabilities.shape == (3, 2, 3)
    assert torch.allclose(probabilities.sum(dim=2), torch.ones(probabilities.shape[0], probabilities.shape[1]), atol=1e-7)

def test_probabilities_from_grouped_data(evaluation: Evaluator):
    # Create synthetic test data groups
    grouped_data = (
        torch.cat(
            [
                torch.randint(0, item_cat, (5, 1))
                for item_cat in evaluation.model.item_categories
            ],
            dim=1,
        ),
        torch.cat(
            [
                torch.randint(0, item_cat, (5, 1))
                for item_cat in evaluation.model.item_categories
            ],
            dim=1,
        ),
        torch.cat(
            [
                torch.randint(0, item_cat, (5, 1))
                for item_cat in evaluation.model.item_categories
            ],
            dim=1,
        ),
    )

    # Call the method
    probabilities = evaluation._grouped_data_probabilities(grouped_data)

    assert probabilities.shape == torch.Size([3, 2, 3])
    assert torch.allclose(probabilities.sum(dim=2), torch.ones_like(probabilities[:, :, 0]), atol=1e-7)

@pytest.mark.parametrize("rescale", [True, False])
def test_latent_group_probabilities(evaluation: Evaluator, rescale):
    # Create some synthetic test data
    data = torch.cat(
        [
            torch.randint(0, item_cat, (30, 1))
            for item_cat in evaluation.model.item_categories
        ],
        dim=1,
    ).float()

    groups = 3
    (
        grouped_data_probabilities,
        grouped_model_probabilities,
        group_averages,
    ) = evaluation.latent_group_probabilities(groups=groups, data=data, rescale=rescale, latent_variable=1)

    # Check the number of groups
    assert grouped_data_probabilities.shape == grouped_model_probabilities.shape
    assert grouped_data_probabilities.shape == torch.Size([3, 2, 3])
    assert torch.allclose(grouped_data_probabilities.sum(dim=2), torch.ones_like(grouped_data_probabilities[:, :, 0]), atol=1e-7)
    assert torch.allclose(grouped_model_probabilities.sum(dim=2), torch.ones_like(grouped_model_probabilities[:, :, 0]), atol=1e-7)
    assert group_averages.shape == (groups,)

    if rescale:
        evaluation.model.transform_theta.assert_called_once()

def test_group_fit_residuals(evaluation: Evaluator):
    data = torch.cat(
        [
            torch.randint(0, item_cat, (30, 1))
            for item_cat in evaluation.model.item_categories
        ],
        dim=1,
    )

    residuals, mid_points = evaluation.group_fit_residuals(data=data, groups=3, latent_variable=1)

    assert residuals.shape == (3, 2, 3)
    assert mid_points.shape == (3,)
    assert evaluation.model.item_probabilities.call_count == 3

@pytest.mark.parametrize(
    "latent_density_method", ["data", "encoder sampling", "qmvn", "gmm"]
)
def test_sum_score_probabilities(evaluation: Evaluator, latent_density_method):
    # mock _cv_gaussian_mixture_model method
    def _cv_gaussian_mixture_model_mock(data, cv_n_components):
        return GaussianMixtureModel(n_components=cv_n_components[0], n_features=data.shape[1])
    
    evaluation._cv_gaussian_mixture_model = MagicMock(side_effect=_cv_gaussian_mixture_model_mock)
    
    if latent_density_method == "encoder sampling":
        with pytest.raises(ValueError):
            total_score_probs = evaluation.sum_score_probabilities(
                latent_density_method=latent_density_method
            )
    else:
        total_score_probs = evaluation.sum_score_probabilities(
            latent_density_method=latent_density_method, trapezoidal_segments=5
        )
        # We should have 4 scores since we have item_categories = [2, 3]
        assert total_score_probs.shape == (4,)
        # They should add up to 1
        assert torch.isclose(total_score_probs.sum(), torch.tensor(1.0), atol=1e-6)

def test_log_likelihood(evaluation: Evaluator):
    # Create synthetic test data
    data = torch.cat(
        [
            torch.randint(0, item_cat, (30, 1))
            for item_cat in evaluation.model.item_categories
        ],
        dim=1,
    )

    # Mock log_likelihood method
    def log_likelihood_mock(data, output, missing_mask, loss_reduction):
        t1 = torch.tensor([-0.5, -1.0]).repeat(data.shape[0] // 2)
        t2 = torch.tensor([-0.1, -0.2]).repeat(data.shape[0] - (data.shape[0] // 2))
        return torch.cat([t1, t2])

    evaluation.model.log_likelihood = MagicMock(side_effect=log_likelihood_mock)

    # latent_group_fit
    # Test with default parameters
    group_mean_likelihoods = evaluation.group_fit_log_likelihood(data=data)
    assert group_mean_likelihoods.shape == (10,)  # Default number of groups is 10
    assert group_mean_likelihoods.dtype == torch.float32
    assert torch.allclose(group_mean_likelihoods[:5], torch.full((5,), -1.5)) # first half of groups
    assert torch.allclose(group_mean_likelihoods[5:], torch.full((5,), -0.3)) # second half of groups
    # Test with specified parameters
    group_mean_likelihoods = evaluation.group_fit_log_likelihood(groups=5, latent_variable=1, data=data, theta_estimation="ML")
    assert group_mean_likelihoods.shape == (5,)  # We specified 5 groups
    assert group_mean_likelihoods.dtype == torch.float32
    assert torch.allclose(group_mean_likelihoods[:2], torch.full((2,), -1.5)) # first half of groups
    assert torch.allclose(group_mean_likelihoods[2], torch.full((1,), -0.9)) # first half of groups
    assert torch.allclose(group_mean_likelihoods[3:], torch.full((2,), -0.3)) # second half of groups

    # respondent level
    # Test with default parameters
    person_likelihoods = evaluation.log_likelihood(data=data, reduction="sum", level="respondent")
    assert person_likelihoods.shape == (30,)  # We have 30 respondents
    assert person_likelihoods.dtype == torch.float32
    assert torch.allclose(person_likelihoods[:15], torch.full((15,), -1.5)) # first half of respondents
    assert torch.allclose(person_likelihoods[15:], torch.full((15,), -0.3)) # second half of respondents

    # item level
    # Test with default parameters
    item_likelihoods = evaluation.log_likelihood(data=data, reduction="mean", level="item")
    assert item_likelihoods.shape == (2,)  # We have 2 items
    assert item_likelihoods.dtype == torch.float32
    assert torch.isclose(item_likelihoods[0], torch.tensor(-0.3))
    assert torch.isclose(item_likelihoods[1], torch.tensor(-0.6))


def test_accuracy(evaluation: Evaluator):
    # Create synthetic test data
    data = torch.cat(
        [
            torch.randint(0, item_cat, (30, 1))
            for item_cat in evaluation.model.item_categories
        ],
        dim=1,
    ).float()
    data[5:7, 1] = torch.nan

    # Mock probabilities
    def item_probabilities_mock(theta):
        # static 3D tensor with dimensions (respondents, items, item categories)
        t1 = torch.tensor(([0.55, 0.45, 0.0], [0.2, 0.35, 0.45])).unsqueeze(0).repeat(theta.shape[0] // 2, 1, 1)
        t2 = torch.tensor(([0.15, 0.85, 0.0], [0.6, 0.05, 0.35])).unsqueeze(0).repeat(theta.shape[0] - (theta.shape[0] // 2), 1, 1)
        return torch.cat([t1, t2])

    evaluation.model.item_probabilities = MagicMock(side_effect=item_probabilities_mock)

    # overall
    accuracy = evaluation.accuracy(data=data, level="all")
    assert accuracy.shape == ()
    assert accuracy.dtype == torch.float32
    assert torch.all((accuracy >= 0) & (accuracy <= 1)), "All values are not between 0 and 1"
    
    # respondent level
    accuracy = evaluation.accuracy(data=data, level="respondent")
    assert accuracy.shape == (30,)  # We have 30 respondents
    assert accuracy.dtype == torch.float32
    assert torch.all((accuracy >= 0) & (accuracy <= 1)), "All values are not between 0 and 1"

    # item level
    item_likelihoods = evaluation.accuracy(data=data, level="item")
    assert item_likelihoods.shape == (2,)  # We have 2 items
    assert item_likelihoods.dtype == torch.float32
    assert torch.all((accuracy >= 0) & (accuracy <= 1)), "All values are not between 0 and 1"


def test_residuals(evaluation: Evaluator):
    # Create synthetic test data
    data = torch.tensor(
        (
            [0., 2.],
            [1., 1.],
            [1., 0.],
            [0., -1.],
            [0., torch.nan],
        )
    )

    # Mock probabilities
    def item_probabilities_mock(theta):
        # static 3D tensor with dimensions (respondents, items, item categories)
        t1 = torch.tensor(([0.55, 0.45, 0.0], [0.2, 0.35, 0.45])).unsqueeze(0).repeat(theta.shape[0] // 2, 1, 1)
        t2 = torch.tensor(([0.15, 0.85, 0.0], [0.6, 0.05, 0.35])).unsqueeze(0).repeat(theta.shape[0] - (theta.shape[0] // 2), 1, 1)
        return torch.cat([t1, t2])

    evaluation.model.item_probabilities = MagicMock(side_effect=item_probabilities_mock)
    # For scored tests, we actual vs model expected scores


    # For MC tests, we have residuals for each MC option
    evaluation.model.mc_correct = [0, 2]
    # non-averaged
    residuals = evaluation.residuals(data=data)
    assert residuals.shape == (5, 2)
    assert torch.allclose(residuals, torch.tensor([
        [0.4500, 0.5500],
        [0.5500, 0.6500],
        [0.1500, 0.4000],
        [0.8500, torch.nan],
        [0.8500, torch.nan]],
    ), equal_nan=True), "Residuals are not correct"

    # overall
    residuals = evaluation.residuals(data=data, average_over="everything")
    assert residuals.shape == ()  # We have 30 respondents
    assert torch.isclose(residuals, torch.tensor(0.556249976)), "Residuals are not correct"

    # respondent level
    residuals = evaluation.residuals(data=data, average_over="items")
    assert residuals.shape == (5,)  # We have 5 respondents
    assert torch.allclose(residuals, torch.tensor([0.5000, 0.6000, 0.2750, 0.8500, 0.8500])), "Residuals are not correct"

    # item level
    residuals = evaluation.residuals(data=data, average_over="respondents")
    assert residuals.shape == (2,)  # We have 2 items
    assert torch.allclose(residuals, torch.tensor([0.5700, 0.53333336114])), "Residuals are not correct"

def test_infit_outfit(evaluation: Evaluator):
    data = torch.tensor([[0, 2], [1, 1]]).float()
    theta = torch.randn(2, evaluation.model.latent_variables)

    def item_probabilities_mock(theta):
        return torch.tensor([[[0.55, 0.45, 0.0], [0.2, 0.35, 0.45]], [[0.15, 0.85, 0.0], [0.6, 0.05, 0.35]]])

    
    evaluation.model.item_probabilities = MagicMock(side_effect=item_probabilities_mock)
    
    def expected_scores_mock(theta, **kwargs):
        return torch.tensor(([0.45, 0.35 + 2 * 0.45], [0.85, 0.05+2 * 0.35]))

    evaluation.model.expected_scores = MagicMock(side_effect=expected_scores_mock)

    infit, outfit = evaluation.infit_outfit(data, theta, level = "item")
    assert torch.allclose(infit, torch.tensor([0.6000000, 0.4237288])), "Infit is incorrect"
    assert torch.allclose(outfit, torch.tensor([0.4973261, 0.5139347])), "Outfit is incorrect"

    infit, outfit = evaluation.infit_outfit(data, theta, level = "respondent")
    assert torch.allclose(infit, torch.tensor([0.9161677, 0.0837438])), "Infit is incorrect"
    assert torch.allclose(outfit, torch.tensor([0.8878143, 0.1234465])), "Outfit is incorrect"

    def expected_scores_mock_mc(theta, **kwargs):
        return torch.tensor(([0.55, 0.45], [0.15, 0.35]))

    evaluation.model.expected_scores = MagicMock(side_effect=expected_scores_mock_mc)

    evaluation.model.mc_correct = [0, 2]
    infit, outfit = evaluation.infit_outfit(data, theta, level = "item")
    assert torch.allclose(infit, torch.tensor([0.6000000, 0.8947369])), "Infit is incorrect"
    assert torch.allclose(outfit, torch.tensor([0.4973262, 0.8803419])), "Outfit is incorrect"

    infit, outfit = evaluation.infit_outfit(data, theta, level = "respondent")
    assert torch.allclose(infit, torch.tensor([1.0202020, 0.4084507])), "Infit is incorrect"
    assert torch.allclose(outfit, torch.tensor([1.0202019, 0.3574660])), "Outfit is incorrect"

    # With missing data
    data[0, 1] = torch.nan
    infit, outfit = evaluation.infit_outfit(data, theta, level = "item")
    assert torch.allclose(infit, torch.tensor([0.6000000, 0.53846150])), "Infit is incorrect"
    assert torch.allclose(outfit, torch.tensor([0.4973262, 0.53846150])), "Outfit is incorrect"
    infit, outfit = evaluation.infit_outfit(data, theta, level = "respondent")
    assert torch.allclose(infit, torch.tensor([0.8181817, 0.4084507]), equal_nan=True), "Infit is incorrect"
    assert torch.allclose(outfit, torch.tensor([0.8181817, 0.3574660]), equal_nan=True), "Outfit is incorrect"

def test_q3(evaluation: Evaluator):
    data = torch.cat(
        [
            torch.randint(0, item_cat, (5, 1))
            for item_cat in evaluation.model.item_categories
        ],
        dim=1,
    ).float()
    data[1:2, 1] = torch.nan

    def residuals_mock(data, theta, average_over, **kwargs):
        residuals = torch.zeros_like(data)
        residuals[:, 0] = 0.5
        residuals[0, 0] = 0
        residuals[:, 1] = torch.arange(data.size()[0]) / data.size()[0]
        residuals[torch.isnan(data)] = torch.nan
        return residuals
    
    evaluation.residuals = MagicMock(side_effect=residuals_mock)

    q3_matrix, _ = evaluation.q3(data=data)

    assert q3_matrix.shape == (len(evaluation.model.item_categories), len(evaluation.model.item_categories)), \
        "q3_matrix does not have the expected shape"
    assert np.allclose(q3_matrix.iloc[0, 1], q3_matrix.iloc[1, 0], atol=1e-3), \
        "q3_matrix is not symmetric"
    assert np.isclose(q3_matrix.iloc[0, 1], 0.878, atol=1e-3), \
        "q3_matrix[0, 1] is not close to the expected value"

def test__observed_item_score_proportions(evaluation: Evaluator):
    data = torch.tensor(
        [
            [1., 2.],
            [1., 0.],
            [torch.nan, 2.],
            [0., 2.],
        ],
    )

    observed_proportions = evaluation._observed_item_score_proportions(data)
    assert torch.allclose(
        observed_proportions,
        torch.tensor([
            [0.3333333, 0.6666667, 0.00],
            [0.2500000, 0.0000000, 0.75]
        ])
    )

@pytest.mark.parametrize("cv_n_components", [[1], [1, 2, 3]])
def test__cv_gaussian_mixture_model(evaluation: Evaluator, cv_n_components):
    data = torch.randn(100, evaluation.model.latent_variables)

    with patch.object(GaussianMixtureModel, "fit") as mock_fit, patch.object(GaussianMixtureModel, "__call__") as mock_call:
        mock_fit.return_value = None
        mock_call.return_value = torch.tensor(1.0)

        gmm = evaluation._cv_gaussian_mixture_model(data, cv_n_components)

    assert isinstance(gmm, GaussianMixtureModel)
    assert gmm.n_components == cv_n_components[0]
    assert gmm.n_features == data.shape[1]
    mock_fit.assert_called()
    if len(cv_n_components) > 1:
        assert mock_call.call_count == len(cv_n_components) * 5  # 5-fold cross-validation


import logging
import copy
import torch
from irtorch.models import BaseIRTModel
from irtorch._internal_utils import dynamic_print, sum_score
from irtorch.irt_dataset import PytorchIRTDataset
from irtorch.estimation_algorithms import BaseIRTAlgorithm

logger = logging.getLogger("irtorch")

class JML(BaseIRTAlgorithm):
    r"""
    Joint Maximum Likelihood (JML) for fitting IRT models :cite:p:`Birnbaum1968`.
    JML optimizes the log-likelihood directly without any latent variable integration or distrbutional assumptions.
    Instead of rotating between optimizing the latent variables and the model parameters as with the typical JML implementation, 
    all parameters are updated at the same time using the Adam optimizer.

    This algorithm is not recommended for large datasets.
    """
    def __init__(
        self,
    ):
        super().__init__()
        self.covariance_matrix = None
        self.optimizer = None
        self.training_history = {
            "train_loss": [],
        }

    def fit(
        self,
        model: BaseIRTModel,
        train_data: torch.Tensor,
        learning_rate: float = 0.1,
        learning_rate_update_patience: int = 80,
        learning_rate_updates_before_stopping: int = 3,
        max_epochs: int = 10000,
        batch_size: int = None,
        start_thetas: torch.Tensor = None,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
    ):
        """
        Train the model.

        Parameters
        ----------
        model : BaseIRTModel
            The model to fit. Needs to inherit :class:`irtorch.models.BaseIRTModel`.
        train_data : torch.Tensor
            The training data. Item responses should be coded 0, 1, ... and missing responses coded as nan or -1.
        learning_rate : float, optional
            The initial learning rate for the optimizer. (default is 0.1)
        learning_rate_update_patience : int, optional
            The number of epochs to wait before reducing the learning rate. (default is 40)
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training. (default is 3)
        max_epochs : int, optional
            The maximum number of epochs to train for. (default is 10000)
        batch_size : int, optional
            The batch size for training. (default is None and uses the full dataset)
        start_thetas : torch.Tensor, optional
            The starting thetas for the training. (default is None and uses the standardized sum scores)
        device : str, optional
            The device to run the model on. (default is "cuda" if available else "cpu".)
        """
        super().fit(model = model, train_data = train_data)

        self.training_history = {
            "train_loss": [],
        }
        if batch_size is None:
            batch_size = train_data.shape[0]

        if start_thetas is None:
            start_thetas = sum_score(train_data, model.mc_correct)
            start_thetas = (start_thetas - start_thetas.mean()) / start_thetas.std()
            start_thetas = start_thetas.detach().unsqueeze(1).repeat(1, model.latent_variables)
            if model.latent_variables > 1:
                # add small noise to avoid all latent variables updating the same way
                start_thetas = start_thetas + torch.randn(start_thetas.shape) * 0.1
        elif start_thetas.shape[0] != train_data.shape[0]:
            raise ValueError("The number of starting thetas must be the same as the number of respondents.")
        elif start_thetas.shape[1] != model.latent_variables:
            raise ValueError("start_thetas must contain the same number of latent variables as the model.")

        self.training_theta_scores = torch.nn.Parameter(
            start_thetas.to(device)
        )

        self.optimizer = torch.optim.Adam(
            list([self.training_theta_scores]) + list(model.parameters()), lr=learning_rate, amsgrad=True
        )

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, mode="min", factor=0.5, patience=learning_rate_update_patience
        )

        train_data_irt = PytorchIRTDataset(
            data=train_data.to(device),
            item_categories=model.item_categories,
            mc_correct=model.mc_correct
        )
        data_loader = torch.utils.data.DataLoader(
            train_data_irt,
            batch_size=batch_size,
            shuffle=False,
            pin_memory=False,
        )

        model.to(device)
        self._training_loop(
            model,
            data_loader,
            scheduler,
            max_epochs,
            learning_rate_updates_before_stopping
        )
        model.to("cpu")
        self.training_theta_scores = self.training_theta_scores.to("cpu")
        model.eval()

    def _training_loop(
        self,
        model: BaseIRTModel,
        data_loader: torch.utils.data.DataLoader,
        scheduler: torch.optim.lr_scheduler.ReduceLROnPlateau,
        max_epochs: int,
        learning_rate_updates_before_stopping: int,
    ):
        """
        The training loop for the model.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        data_loader : torch.utils.data.DataLoader
            The data loader for the training data.
        scheduler : torch.optim.lr_scheduler.ReduceLROnPlateau
            The learning rate scheduler.
        max_epochs : int
            The maximum number of epochs to train for.
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training.
        """
        lr_update_count = 0

        best_loss = float("inf")
        best_model_state = None
        best_epoch = 0
        prev_lr = [group["lr"] for group in self.optimizer.param_groups]
        model.train()

        for epoch in range(max_epochs):
            for _, (batch, mask, _, idx) in enumerate(data_loader):
                self.optimizer.zero_grad()
                model_out = model(self.training_theta_scores[idx])
                loss = -model.log_likelihood(batch, model_out, missing_mask=mask, loss_reduction="sum")
                loss.backward()
                self.optimizer.step()
                
                train_loss = loss.item()
                self.training_history["train_loss"].append(loss.item())
                
                current_loss = train_loss
                scheduler.step(train_loss)
                dynamic_print(f"Epoch: {epoch}. Loss: {train_loss:.4f}")

                if current_loss < best_loss:
                    best_loss = current_loss
                    best_epoch = epoch
                    best_model_state = {
                        "theta": copy.deepcopy(self.training_theta_scores.data),
                        "state_dict": copy.deepcopy(model.state_dict()),
                        "optimizer": copy.deepcopy(self.optimizer.state_dict()) 
                    }

                current_lr = [group["lr"] for group in self.optimizer.param_groups]
                if current_lr != prev_lr:
                    lr_update_count += 1
                    if lr_update_count >= learning_rate_updates_before_stopping:
                        break
                    prev_lr = current_lr
                    logger.info("Decreased learning rate to: %s", current_lr[0])


            if lr_update_count >= learning_rate_updates_before_stopping:
                break

        # Load the best model state
        if best_model_state is not None:
            self.training_theta_scores = best_model_state["theta"]
            model.load_state_dict(best_model_state["state_dict"])
            self.optimizer.load_state_dict(best_model_state["optimizer"])
            logger.info("Best model found at iteration %s with loss %.4f.", best_epoch, best_loss)



import torch
from torch import nn

class NegationLayer(nn.Module):
    def __init__(self, item_theta_relationships: torch.Tensor, inputs_per_item: int, zero_outputs: torch.Tensor):
        """
        Parameters
        ----------
        item_theta_relationships : torch.Tensor
            An integer tensor of shape (items,) that determines for which weights should be trained. Mostly used for training multidimensional IRT models for which each item or item category has a separate network and some of the item/latent variable relationships are set to be 0.
        inputs_per_item : int
            Number of inputs per item. One weight for all inputs per item.
        zero_outputs : torch.Tensor
            A boolean tensor of shape (out_features,) that determines which outputs should be 0. Typically used for the final layer for IRT models with polytomously scored items where different items have differing number of response categories. Items with fewer categories should set the outputs for nonexisting categories to 0.
        """
        super().__init__()

        self.inputs_per_item = inputs_per_item
        self.item_theta_relationships = item_theta_relationships
        self.zero_outputs = zero_outputs
        self.zero_weights = (1-item_theta_relationships).repeat_interleave(inputs_per_item).bool()
        self.register_buffer("weight", torch.zeros(item_theta_relationships.numel() * inputs_per_item))
        # One weight for each item related to the latent variable
        self.weight_param = nn.Parameter(torch.Tensor(sum(item_theta_relationships).item()))
        self.reset_parameters()

    def reset_parameters(self):
        # nn.init.zeros_(self.weight_param)
        nn.init.constant_(self.weight_param, .01)

    def forward(self, x: torch.Tensor):
        # break the gradients for weights for non-used categories
        x[:, self.zero_outputs] = 0.0
        weight = self.weight.clone().detach()
        weight[~self.zero_weights] = self.weight_param.repeat_interleave(self.inputs_per_item)
        x.multiply_(weight)
        return x

    def all_item_weights(self):
        """
        Returns
        -------
        weight : torch.Tensor
            All layer weights, including the unused ones.
        """
        weights = torch.zeros(self.item_theta_relationships.numel())
        mask = (1 - self.item_theta_relationships).bool()
        weights[~mask] = self.weight_param
        return weights



import pytest
from unittest.mock import MagicMock
import torch
from irtorch.utils import gauss_hermite, get_item_categories, split_data, impute_missing, _imw_newton_raphson, imw

def test_gauss_hermite():
    n_points = 4
    mu = torch.tensor([1, 0])
    Sigma = torch.tensor([[1.3, -0.213], [-0.213, 1.2]])
    points, weights = gauss_hermite(n_points, mu, Sigma)

    # Calculations
    mean = torch.sum((weights)[:, None] * points, dim=0)
    variance = torch.sum((weights)[:, None] * (points - mu)**2, dim=0)
    # Covariance calculation using outer product
    covfunc = lambda x: torch.ger(x - mu, x - mu)
    covariance = torch.sum((weights)[:, None, None] * torch.stack(list(map(covfunc, points))), dim=0)
    # Polynomial calculation
    polynomial = torch.sum((weights)[:, None] * (4 + 2 * points - 0.5 * points**2), dim=0)
    # Logistic function calculation
    logistic = torch.sum((weights)[:, None] * (1 / (1 + torch.exp(-points))), dim=0)
    # Assertions
    assert torch.allclose(mean, torch.tensor([1.0000, 0.0000], dtype=torch.float64), atol=1e-4)
    assert torch.allclose(variance, torch.tensor([1.3000, 1.2000], dtype=torch.float64), atol=1e-4)
    assert torch.allclose(covariance, torch.tensor([[1.3000, -0.2130], [-0.2130, 1.2000]], dtype=torch.float64), atol=1e-4)
    assert torch.allclose(polynomial, torch.tensor([4.8500, 3.4000], dtype=torch.float64), atol=1e-4)
    assert torch.allclose(logistic, torch.tensor([0.6888, 0.5000], dtype=torch.float64), atol=1e-4)


def test_get_item_categories(test_data, item_categories):
    result = get_item_categories(test_data)
    assert result == item_categories

def test_impute_missing():
    data = torch.tensor([[1, 2, 1, -1], [-1, float("nan"), 0, 2], [-1, float("nan"), 1, 2]])
    imputed_data = impute_missing(data = data, method="zero")
    assert (imputed_data == torch.tensor([[1, 2, 1, 0], [0, 0, 0, 2], [0, 0, 1, 2]])).all()

    imputed_data = impute_missing(data = data, method="mean")
    assert (imputed_data == torch.tensor([[1, 2, 1, 2], [1, 2, 0, 2], [1, 2, 1, 2]])).all()

    mc_correct = [1, 2, 1, 2]
    imputed_data = impute_missing(data = data, method = "random incorrect", mc_correct=mc_correct, item_categories=[3, 3, 2, 3])
    missing_mask = torch.logical_or(data == -1, data.isnan())
    assert (imputed_data[missing_mask] > -1).all() # all missing are replaced
    assert (imputed_data[~missing_mask] == data[~missing_mask]).all() # non missing are still the same
    assert imputed_data[0, 3] != 2 # we did not replace with true values
    assert imputed_data[1, 0] != 1
    assert imputed_data[1, 1] != 2
    assert imputed_data[2, 0] != 1
    assert imputed_data[2, 1] != 2


def test__imw_newton_raphson():
    # Negative entropy values
    negative_entropy = torch.tensor([-0.4, -0.6], requires_grad=False)
    w = _imw_newton_raphson(negative_entropy)
    negative_entropy_result = w * torch.log(w) + (1 - w) * torch.log(1 - w)
    assert torch.allclose(negative_entropy_result, negative_entropy, atol=1e-6), "The computed w does not satisfy the equation."

def test_imw():
    # Mock models
    model1 = MagicMock()
    model2 = MagicMock()
    model1.evaluate.log_likelihood = MagicMock(return_value=torch.tensor(-0.5))
    model2.evaluate.log_likelihood = MagicMock(return_value=torch.tensor(-0.2))
    # Data (can be empty since we are mocking)
    data = torch.tensor([])
    # Call imw function
    imw_value = imw(model1, model2, data)
    # Compute expected imw
    log_geometric_means = torch.tensor([-0.5, -0.2])
    w = _imw_newton_raphson(log_geometric_means)
    expected_imw = (w[1] - w[0]) / w[0]
    # Check that imw_value matches expected_imw
    assert torch.allclose(imw_value, expected_imw, atol=1e-6), "The computed IMW does not match the expected value."

def test_split_data():
    torch.manual_seed(42)
    data = torch.rand(100, 10)

    # Test with shuffle=True
    train_data, test_data = split_data(data, train_ratio=0.8, shuffle=True)
    assert train_data.shape[0] == 80  # 80% of 100 samples
    assert test_data.shape[0] == 20  # 20% of 100 samples
    assert not torch.equal(data[:80], train_data)  # Data should be shuffled

    # Test with shuffle=False
    train_data, test_data = split_data(data, train_ratio=0.8, shuffle=False)
    assert train_data.shape[0] == 80
    assert test_data.shape[0] == 20
    assert torch.equal(data[:80], train_data)  # Data should not be shuffled

    # Test with different train_ratio
    train_data, test_data = split_data(data, train_ratio=0.6, shuffle=False)
    assert train_data.shape[0] == 60
    assert test_data.shape[0] == 40

    # Test with train_ratio=1.0
    train_data, test_data = split_data(data, train_ratio=1.0, shuffle=False)
    assert train_data.shape[0] == 100
    assert test_data.shape[0] == 0

    # Test with train_ratio=0.0
    train_data, test_data = split_data(data, train_ratio=0.0, shuffle=False)
    assert train_data.shape[0] == 0
    assert test_data.shape[0] == 100



import pandas as pd
import torch
import torch.nn as nn
from irtorch.models.base_irt_model import BaseIRTModel

class GradedResponse(BaseIRTModel):
    r"""
    Graded response IRT model :cite:p:`Samejima1968`.

    Parameters
    ----------
    data: torch.Tensor, optional
        A 2D torch tensor with test data. Used to automatically compute item_categories. Columns are items and rows are respondents. (default is None)
    latent_variables : int
        Number of latent variables.
    item_categories : list[int]
        Number of categories for each item. One integer for each item. Missing responses exluded.
    item_theta_relationships : torch.Tensor, optional
        A boolean tensor of shape (items, latent_variables). If specified, the model will have connections between latent dimensions and items where the tensor is True. If left out, all latent variables and items are related (Default: None)

    Notes
    -----
    For an item :math:`j` with ordered item scores :math:`x=0, 1, 2, ...` the model defines the probability for responding with a score of :math:`x` or higher for all :math:`x>0` as follows:

    .. math::

        P(X_j \geq x | \mathbf{\theta}) = \dfrac{\exp \left(\mathbf{a}_{j}^\top \mathbf{\theta} + d_{jx}\right)}{1+\exp \left(\mathbf{a}_{j}^\top \mathbf{\theta} + d_{jx}\right)}

            
    where:

    - :math:`\mathbf{\theta}` is a vector of latent variables.
    - :math:`\mathbf{a}_{j}` is a vector of weights for item :math:`j`.
    - :math:`d_{jx}` is the bias term for item :math:`j` and score :math:`x`.

    From here, the probability of responding with a score of :math:`x` is calculated as:

    .. math::
        P(X_j = x | \mathbf{\theta}) = \begin{cases}
            1-P(X_j \geq x +1 | \mathbf{\theta}), & \text{if } x = 0\\
            P(X_j \geq x | \mathbf{\theta})-P(X_j \geq x+1 | \mathbf{\theta}), & \text{otherwise}
        \end{cases}

    Examples
    --------
    >>> from irtorch.models import GradedResponse
    >>> from irtorch.estimation_algorithms import AE
    >>> from irtorch.load_dataset import swedish_national_mathematics_1
    >>> data = swedish_national_mathematics_1()
    >>> model = GradedResponse(data)
    >>> model.fit(train_data=data, algorithm=AE())
    """
    def __init__(
        self,
        data: torch.Tensor = None,
        latent_variables: int = 1,
        item_categories: list[int] = None,
        item_theta_relationships: torch.Tensor = None,
    ):
        if item_categories is None:
            if data is None:
                raise ValueError("Either item_categories or data must be provided to initialize the model.")
            else:
                # replace nan with -inf to get max
                item_categories = (torch.where(~data.isnan(), data, torch.tensor(float('-inf'))).max(dim=0).values + 1).int().tolist()

        super().__init__(latent_variables=latent_variables, item_categories=item_categories)
        if item_theta_relationships is not None:
            if item_theta_relationships.shape != (len(item_categories), latent_variables):
                raise ValueError(
                    f"latent_item_connections must have shape ({len(item_categories)}, {latent_variables})."
                )
            assert(item_theta_relationships.dtype == torch.bool), "latent_item_connections must be boolean type."
            assert(torch.all(item_theta_relationships.sum(dim=1) > 0)), "all items must have a relationship with a least one latent variable."

        self.output_size = self.items * self.max_item_responses

        free_weights = torch.ones(self.items, latent_variables)
        if item_theta_relationships is not None:
            for item, item_cat in enumerate(self.item_categories):
                free_weights[item, :] = item_theta_relationships[item, :]

        self.weight_param = nn.Parameter(torch.zeros(free_weights.sum().int()))
        
        first_category = torch.zeros(self.items, self.max_item_responses)
        first_category[:, 0] = 1.0
        first_category = first_category.reshape(-1)
        missing_category = torch.zeros(self.items, self.max_item_responses)
        for item, item_cat in enumerate(self.item_categories):
            missing_category[item, item_cat:self.max_item_responses] = 1.0
        missing_category = missing_category.reshape(-1)
        free_bias = (1 - first_category) * (1 - missing_category)

        self.register_buffer("free_weights", free_weights.bool())
        self.register_buffer("free_bias", free_bias.bool())
        self.register_buffer("missing_category", missing_category.bool())
        self.register_buffer("first_category", first_category.bool())
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Xavier uniform initialization for weights https://pytorch.org/docs/stable/nn.init.html#torch.nn.init.xavier_uniform_
        a = torch.sqrt(torch.tensor(6 / (self.latent_variables + self.items)))
        nn.init.uniform_(self.weight_param, a=-a, b=a)
        # initialize bias parameters
        bias = torch.zeros_like(self.free_bias, dtype=torch.float)
        for item in range(self.items):
            item_bias = torch.zeros(self.max_item_responses, dtype=torch.float)
            item_ind = range(item * self.max_item_responses, item * self.max_item_responses + self.max_item_responses)
            probabilities = torch.linspace(
                1/(1+torch.exp(torch.tensor(4))),
                1/(1+torch.exp(torch.tensor(-4))),
                self.free_bias[item_ind].sum() + 2
            )[1:-1]
            item_bias[self.free_bias[item_ind]] = probabilities.pow(-1).add(-1).log()
            bias[item_ind] = item_bias

        self.bias_param = nn.Parameter(bias[self.free_bias])
        # initial_bias = -torch.arange(-1., 1.01, 2/(self.max_item_responses - 1)).tile((self.items, 1)).flatten()

    def forward(self, theta: torch.Tensor) -> torch.Tensor:
        r"""
        Forward pass of the model.

        Parameters
        ----------
        theta : torch.Tensor
            2D tensor with latent variables. Rows are respondents and latent variables are columns. 

        Returns
        -------
        output : torch.Tensor
            2D tensor. Rows are respondents and columns are :math:`\mathbf{a}_{j}^\top \mathbf{\theta} + d_{jx}`.
        """
        bias = torch.zeros(self.output_size, device=theta.device)
        bias[self.free_bias] = self.bias_param
        
        weights = torch.zeros(self.items, self.latent_variables, device=theta.device)
        weights[self.free_weights] = self.weight_param
        weighted_theta = torch.matmul(theta, weights.T).repeat_interleave(self.max_item_responses, dim=1)

        output = weighted_theta + bias
        output[:, self.missing_category] = -torch.inf
        output[:, self.first_category] = torch.inf

        return output

    def probabilities_from_output(self, output: torch.Tensor) -> torch.Tensor:
        """
        Compute probabilities from the output tensor from the forward method.

        Parameters
        ----------
        output : torch.Tensor
            Output tensor from the forward pass.

        Returns
        -------
        torch.Tensor
            3D tensor with dimensions (respondents, items, item categories)
        """
        # great than or equal to probabilities as per graded response model
        geq_probs = output.sigmoid().reshape(output.shape[0], self.items, self.max_item_responses)
        # convert to probabilities for each score
        probs = geq_probs.clone()  # Create a copy of `probs` to avoid modifying the original in place
        probs[:, :, :-1] -= geq_probs[:, :, 1:]
        return probs

    def log_likelihood(
        self,
        data: torch.Tensor,
        output: torch.Tensor,
        missing_mask: torch.Tensor = None,
        loss_reduction: str = "sum",
    ) -> torch.Tensor:
        """
        Compute the log likelihood.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents
        output: torch.Tensor
            A 2D tensor with output. Columns are item response categories and rows are respondents
        missing_mask: torch.Tensor, optional
            A 2D tensor with missing data mask. (default is None)
        loss_reduction: str, optional 
            The reduction argument. (default is 'sum')
        
        Returns
        -------
        torch.Tensor
            The log likelihood.
        """
        probabilities = self.probabilities_from_output(output)
        data = data.long().view(-1)
        reshaped_probabilities = probabilities.reshape(-1, self.max_item_responses)

        if missing_mask is not None:
            missing_mask = missing_mask.view(-1)
            reshaped_probabilities = reshaped_probabilities[~missing_mask]
            respondents = data.size(0)
            data = data[~missing_mask]

        ll = reshaped_probabilities[torch.arange(data.size(0)), data].log()
        if loss_reduction == "sum":
            return ll.sum()
        elif loss_reduction == "none" and missing_mask is not None:
            ll_masked = torch.full((respondents, ), torch.nan, device= ll.device)
            ll_masked[~missing_mask] = ll
            return ll_masked
        else:
            raise ValueError("loss_reduction must be 'sum' or 'none'")

    def item_parameters(self, irt_format = False) -> pd.DataFrame:
        """
        Get the item parameters for a fitted model.

        Parameters
        ----------
        irt_format : bool, optional
            Only for unidimensional models. Whether to return the item parameters in traditional IRT format. Otherwise returns weights and biases. (default is False)

        Returns
        -------
        pd.DataFrame
            A dataframe with the item parameters.
        """
        biases = torch.zeros(self.output_size)
        biases[self.free_bias] = self.bias_param
        biases = biases.reshape(-1, self.max_item_responses)

        weights = torch.zeros(self.items, self.latent_variables)
        weights[self.free_weights] = self.weight_param

        weights_df = pd.DataFrame(weights.detach().numpy())
        weights_df.columns = [f"a{i+1}" for i in range(weights.shape[1])]
        if irt_format and self.latent_variables == 1:
            biases_df = pd.DataFrame(-(biases/weights).detach()[:, 1:].numpy())
            biases_df.columns = [f"b{i+1}" for i in range(biases_df.shape[1])]
        else:
            biases_df = pd.DataFrame(biases.detach()[:, 1:].numpy())
            biases_df.columns = [f"d{i+1}" for i in range(biases_df.shape[1])]
            
        parameters = pd.concat([weights_df, biases_df], axis=1)

        return parameters

    @torch.no_grad()
    def item_theta_relationship_directions(self, theta:torch.Tensor = None) -> torch.Tensor:
        """
        Get the relationship directions between each item and latent variable for a fitted model.

        Parameters
        ----------
        theta : torch.Tensor, optional
            Not needed for this model. (default is None)

        Returns
        -------
        torch.Tensor
            A 2D tensor with the relationships between the items and latent variables. Items are rows and latent variables are columns.
        """
        weights = torch.zeros(self.items, self.latent_variables)
        weights[self.free_weights] = self.weight_param
        return weights.sign().int()
    


from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from factor_analyzer import Rotator
import torch
from irtorch.rescale import Scale

if TYPE_CHECKING:
    from irtorch.models.base_irt_model import BaseIRTModel

logger = logging.getLogger("irtorch")

class Rotate(Scale):
    """
    Rotates the latent variables to improve interpretability. Utilizes the factor_analyzer package for rotations <https://pypi.org/project/factor-analyzer/>
    
    If the model has already been rescaled using :meth:`irtorch.rescale` the rotation is applied to the rescaled latent variables.
    If you do not want this, use :meth:`irtorch.models.BaseIRTModel.detach_rescale` before applying the rotation.
    
    Parameters
    ----------
    model : BaseIRTModel
        The IRT model which scales to rotate.
    data : torch.Tensor, optional
        The popluation data to compute the latent variable "loadings" for each item.
    theta : torch.Tensor, optional
        The original scale theta scores to compute the latent variable "loadings" for each item.
    loadings: torch.Tensor, optional
        A torch tensor with the loadings for each item. If specified, data and theta are ignored and the loadings are used for rotation. (default is
    rotation_method : str, optional
        The rotation method to use. For available options, see . (default is "promax")
    rotation_matrix : torch.Tensor, optional
        A torch tensor with the rotation matrix. If specified, data, theta and rotation_method are ignored
        and the rotation matrix is used directly. (default is None)
    **kwargs
        Additional keyword arguments used for theta estimation. Refer to :meth:`irtorch.models.BaseIRTModel.latent_scores` for additional details.
    """
    def __init__(
        self,
        model: BaseIRTModel,
        data: torch.Tensor = None,
        theta: torch.Tensor = None,
        loadings: torch.Tensor = None,
        rotation_method: str = "promax",
        rotation_matrix: torch.Tensor = None,
        **kwargs
    ):
        super().__init__(invertible=True)
        if rotation_matrix is None:
            if loadings is None:
                if theta is None:
                    if data is None:
                        raise ValueError("data, theta, loadings or a rotation matrix must be provided for rotation.")
                    theta = model.latent_scores(data, rescale=True ,**kwargs)
                loadings = model.expected_item_score_gradients(theta).mean(dim=0)
            rotator = Rotator(method = rotation_method)
            self.rot_loadings = torch.from_numpy(rotator.fit_transform(loadings.numpy())).float()
            self.rotation_matrix = torch.tensor(rotator.rotation_, dtype=torch.float)
        else:
            if rotation_matrix.shape[0] != model.latent_variables or rotation_matrix.shape[1] != model.latent_variables:
                raise ValueError("The rotation matrix must have the same dimensions as the number of latent variables.")
            self.rotation_matrix = rotation_matrix

        self.inverse_rotation_matrix = torch.inverse(self.rotation_matrix)

    def transform(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Transforms the input theta scores into the new scale.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.
        """
        return torch.matmul(theta, self.rotation_matrix)

    def inverse(self, transformed_theta: torch.Tensor) -> torch.Tensor:
        """
        Puts the scores back to the original theta scale.

        Parameters
        ----------
        transformed_theta : torch.Tensor
            A 2D tensor containing transformed theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A 2D tensor containing theta scores on the the original scale.
        """
        unrotated_theta = torch.matmul(transformed_theta, self.inverse_rotation_matrix)
        return unrotated_theta


    def jacobian(
        self,
        theta: torch.Tensor
    ) -> torch.Tensor:
        r"""
        Computes the gradients of rotated scores for each :math:`j` with respect to the original theta scores.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores on the original theta scale. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A torch tensor with the gradients for each theta score. Dimensions are (theta rows, latent variables, latent variables) where the last two are the jacobians.
        """
        jacobians = self.rotation_matrix.T.unsqueeze(0).repeat(theta.shape[0], 1, 1)
        return jacobians



from abc import ABC, abstractmethod
import torch

class Scale(ABC):
    """
    Abstract base class for Item Response Theory model scale transformations.
    All scale transformations should inherit from this class.

    Not that you can make custom transformations by inheriting from this class.
    A class instance can then be supplied to :meth:`irtorch.models.BaseIRTModel.rescale` to apply the transformation to the latent variables of the model.
    The gradients method is not needed for 
    """
    def __init__(self, invertible: bool = False):
        self.invertible = invertible

    def __call__(self, *args, **kwargs):
        return self.transform(*args, **kwargs)
    
    @abstractmethod
    def transform(self, theta: torch.Tensor) -> torch.Tensor:
        """
        Transforms the input theta scores into the new scale.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing transformed theta scores. Each column represents one latent variable.
        """

    @abstractmethod
    def inverse(self, transformed_theta: torch.Tensor) -> torch.Tensor:
        """
        Puts the scores back to the original theta scale.

        Parameters
        ----------
        transformed_theta : torch.Tensor
            A 2D tensor containing transformed theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A 2D tensor containing theta scores on the the original scale.
        """

    @abstractmethod
    def jacobian(
        self,
        theta: torch.Tensor
    ) -> torch.Tensor:
        r"""
        Computes the Jacobian matrix of the scale transformations for each row in the input theta scores.

        Parameters
        ----------
        theta : torch.Tensor
            A 2D tensor containing latent variable theta scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A torch tensor with the Jacobians for each theta score. Dimensions are (theta rows, latent variables, latent variables) where the last two are the jacobians for each row.
        """



import logging
import copy
import torch
from torch.distributions import MultivariateNormal
from irtorch.models import BaseIRTModel
from irtorch.utils import gauss_hermite
from irtorch._internal_utils import dynamic_print
from irtorch.irt_dataset import PytorchIRTDataset
from irtorch.estimation_algorithms import BaseIRTAlgorithm

logger = logging.getLogger("irtorch")

class MML(BaseIRTAlgorithm):
    r"""
    Marginal Maximum Likelihood (MML) for fitting IRT models :cite:p:`Bock1981`. 
    Uses a multivariate normal distribution for the latent variables and Gradient Descent to optimize the model parameters.
    This method is generally effecive for models with a small number of latent variables. More than 3 is not supported.
    Note that this method runs much faster on a GPU.

    The marginal log-likelihood is calculated by integrating over an assumed normal distribution for the latent variables with density :math:`f(\mathbf{\theta})`.

    .. math::

        \log L(\phi) = \sum_{i=1}^N \log \left( \int P(\mathbf{X}_i = \mathbf{x}_i | \mathbf{\theta}, \phi)f(\mathbf{\theta})d\mathbf{\theta} \right)

    where

    - :math:`N` are the number of respondents,
    - :math:`\mathbf{X}_i` is the response vector of the :math:`i`-th respondent,
    - :math:`\mathbf{x}_i` is the observed response vector of the :math:`i`-th respondent,
    - :math:`\phi` are the model parameters,
    - :math:`\mathbf{\theta}` is the latent variable vector,

    The integral is approximated using Gauss-Hermite quadratures or a Quasi-Monte Carlo method. 
    :math:`\log L(\phi)` is then maximized using stochastic gradient descent. These steps are repeated until convergence.
    """
    def __init__(
        self,
    ):
        super().__init__()
        self.covariance_matrix = None
        self.optimizer = None
        self.training_history = {
            "train_loss": [],
        }

    def fit(
        self,
        model: BaseIRTModel,
        train_data: torch.Tensor,
        max_epochs: int = 1000,
        integration_method: str = "quasi_mc",
        quadrature_points: int = None,
        covariance_matrix: torch.Tensor = None,
        learning_rate: float = 0.15,
        learning_rate_update_patience: int = 7,
        learning_rate_updates_before_stopping: int = 2,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
    ):
        """
        Train the model.

        Parameters
        ----------
        model : BaseIRTModel
            The model to fit. Needs to inherit :class:`irtorch.models.BaseIRTModel`.
        train_data : torch.Tensor
            The training data. Item responses should be coded 0, 1, ... and missing responses coded as nan or -1.
        max_epochs : int, optional
            The maximum number of epochs to train for. (default is 1000)
        integration_method : str, optional
            The method to use for approximating integrals over the latent variables. Can be either "gauss_hermite" for Gauss-Hermite quadrature
            or "quasi_mc" for quasi-Monte Carlo. (default is "quasi_mc").
        quadrature_points : int, optional
            The number of quadrature points to use for latent variable integration. Note that large datasets may lead to memory issues if quadratures points are too high. (default is 'None' and uses a function of the number of latent variables)
        covariance_matrix : torch.Tensor, optional
            The covariance matrix for the multivariate normal distribution for the latent variables. (default is None and uses uncorrelated variables)
        learning_rate : float, optional
            The initial learning rate for the optimizer. (default is 0.15)
        learning_rate_update_patience : int, optional
            The number of epochs to wait before reducing the learning rate. (default is 7)
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training. (default is 2)
        device : str, optional
            The device to run the model on. (default is "cuda" if available else "cpu".)
        """
        super().fit(model = model, train_data = train_data)

        self.training_history = {
            "train_loss": [],
        }

        if quadrature_points is None:
            if model.latent_variables > 3:
                raise ValueError("MML is not implemented for models with more than 3 latent variables because of large integration grid.")
            quadrature_points = {
                1: 15,
                2: 7,
                3: 5
            }.get(model.latent_variables, 20)

        if covariance_matrix is not None:
            if covariance_matrix.shape[0] != model.latent_variables or covariance_matrix.shape[1] != model.latent_variables:
                raise ValueError("Covariance matrix must have the same dimensions as the latent variables.")
            self.covariance_matrix = covariance_matrix
        else:
            self.covariance_matrix = torch.eye(model.latent_variables)

        self.optimizer = torch.optim.Adam(
            list(model.parameters()), lr=learning_rate, amsgrad=True
        )

        if integration_method == "gauss_hermite":
            points, weights = gauss_hermite(
                n=quadrature_points,
                mean=torch.zeros(model.latent_variables),
                covariance=self.covariance_matrix
            )
            log_weights = weights.log()
        else:
            points, log_weights = self._quasi_mc(n_points=quadrature_points, latent_variables=model.latent_variables)

        # Reduce learning rate when loss stops decreasing ("min")
        # we multiply the learning rate by the factor
        # patience: We need no improvement after x epochs for it to trigger
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, mode="min", factor=0.6, patience=learning_rate_update_patience
        )

        model.to(device)
        self._training_loop(
            model,
            train_data.to(device),
            max_epochs,
            points.to(device, dtype = torch.float32),
            log_weights.to(device, dtype = torch.float32),
            scheduler,
            learning_rate_updates_before_stopping
        )
        model.to("cpu")
        model.eval()

    def _training_loop(
        self,
        model: BaseIRTModel,
        train_data: torch.Tensor,
        max_epochs: int,
        points: torch.Tensor,
        log_weights: torch.Tensor,
        scheduler: torch.optim.lr_scheduler.ReduceLROnPlateau,
        learning_rate_updates_before_stopping: int,
    ):
        """
        The training loop for the model.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        train_data : torch.Tensor
            The training data.
        max_epochs : int
            The maximum number of epochs to train for.
        points : torch.Tensor
            The latent variable points to evaluate the MML integral.
        log_weights : torch.distributions.MultivariateNormal
            The logarithm of the integral weights associated with the points.
        scheduler : torch.optim.lr_scheduler.ReduceLROnPlateau
            The learning rate scheduler.
        learning_rate_updates_before_stopping : int, optional
            The number of times the learning rate can be reduced before stopping training.
        """
        lr_update_count = 0

        latent_combos_rep = points.repeat_interleave(train_data.size(0), dim=0)
        train_data_rep = train_data.repeat(points.size(0), 1)
        log_weights_rep = log_weights.repeat_interleave(train_data.size(0), dim=0)
        irt_dataset_rep = PytorchIRTDataset(data=train_data_rep)

        best_loss = float("inf")
        prev_lr = [group["lr"] for group in self.optimizer.param_groups]
        for epoch in range(max_epochs):
            train_loss = self._train_step(
                model,
                irt_dataset_rep,
                latent_combos_rep,
                log_weights_rep,
                points.size(0)
            )

            current_loss = train_loss
            scheduler.step(train_loss)
            dynamic_print(f"Epoch: {epoch}. Loss: {train_loss:.4f}")

            if current_loss < best_loss:
                best_loss = current_loss
                best_epoch = epoch
                best_model_state = { "state_dict": copy.deepcopy(model.state_dict()),
                                    "optimizer": copy.deepcopy(self.optimizer.state_dict()) }
            

            current_lr = [group["lr"] for group in self.optimizer.param_groups]
            # Check if the learning rate has been updated
            if current_lr != prev_lr:
                lr_update_count += 1
                prev_lr = current_lr

            if lr_update_count >= learning_rate_updates_before_stopping:
                logger.info("Stopping training after %s learning rate updates.", learning_rate_updates_before_stopping)
                break

            logger.debug("Current learning rate: %s", self.optimizer.param_groups[0]["lr"])

        # Load the best model state
        if best_model_state is not None:
            model.load_state_dict(best_model_state["state_dict"])
            self.optimizer.load_state_dict(best_model_state["optimizer"])
            logger.info("Best model found at iteration %s with loss %.4f.", best_epoch, best_loss)

    def _train_step(
        self,
        model: BaseIRTModel,
        train_data: PytorchIRTDataset,
        latent_grid: torch.Tensor,
        log_weights: torch.Tensor,
        number_of_weights: int,
    ):
        """
        Training step for an epoch.

        Parameters
        ----------
        model : BaseIRTModel
            The model to train.
        train_data : PytorchIRTDataset
            The training data.
        latent_grid : torch.Tensor
            The grid of latent variables.
        log_weights : torch.Tensor
            The log weights for the latent variables.
        number_of_weights : int
            The number of quadrature points. The number of different weights before expanding.

        Returns
        -------
        float
            The loss after the training step.
        """
        model.train()

        self.optimizer.zero_grad()
        logits = model(latent_grid)
        ll = model.log_likelihood(train_data.data, logits, missing_mask=train_data.mask, loss_reduction="none")
        ll = ll.view(-1, model.items).nansum(dim=1) # sum over items
        
        log_sums = (log_weights + ll).view(number_of_weights, -1)
        constant = log_sums.max(dim=0)[0] # for logexpsum trick (one constant per respondent)
        exp_log_sums = (log_sums-constant).exp()
        loss = -(exp_log_sums.sum(dim=0).log() + constant).sum()

        loss.backward()
        self.optimizer.step()
        
        self.training_history["train_loss"].append(loss.item())
        return loss.item()

    def _quasi_mc(self, n_points: int, latent_variables: int) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Get the points and weights for Quasi-Monte Carlo integral approximation.

        Parameters
        ----------
        n_points : int
            The number of points of evaluation.
        latent_variables : int
            The number of latent variables.

        Returns
        -------
        torch.Tensor
            The points of integration.
        torch.Tensor
            The logarithm of the weights associated with the points.
        """
        latent_grid = torch.linspace(-3, 3, n_points).view(-1, 1)
        latent_grid = latent_grid.expand(-1, latent_variables).contiguous()
        if latent_variables > 1:
            columns = [latent_grid[:, i] for i in range(latent_grid.size(1))]
            latent_combos = torch.cartesian_prod(*columns)
        else:
            latent_combos = latent_grid

        normal_dist = MultivariateNormal(
            loc=torch.zeros(latent_variables),
            covariance_matrix=self.covariance_matrix
        )
        weights = normal_dist.log_prob(latent_combos)
        return latent_combos, weights



import torch
from torch import nn
import torch.nn.functional as F

class MonotonePolynomialModule(nn.Module):
    """
    A polynomial with monotonicity constraints.

    Parameters
    ----------
    degree: int
        Degree of the polynomial. Needs to be an uneven number.
    in_features: int
        Number of input features.
    out_features: int
        Number of output features.
    intercept: bool
        Whether to include an intercept term. (Default: False)
    relationship_matrix : torch.Tensor, optional
        A boolean tensor of shape (in_features, out_features,) that determines which inputs are related to which outputs. Typically used for IRT models to remove relationships between some items or item categories and latent variables. (Default: None)
    negative_relationships : bool, optional
        Whether to allow for negative relationships. (Default: False)
    shared_directions : int, optional
        Only when negative_relationships is true. Number of out_features with shared relationship directions. out_features needs to be divisible with this. (Default: 1)
    """
    def __init__(
        self,
        degree: int,
        in_features: int = 1,
        out_features: int = 1,
        intercept: int = False,
        relationship_matrix: torch.Tensor = None,
        negative_relationships: bool = False,
        shared_directions: int = 1
    ) -> None:
        super().__init__()
        if degree % 2 == 0:
            raise ValueError("Degree must be an uneven number.")
        self.k = (degree - 1) // 2
        self.input_dim = in_features
        self.output_dim = out_features
        self.relationship_matrix = relationship_matrix
        self.negative_relationships = negative_relationships
        self.shared_directions = shared_directions

        self.omega = nn.Parameter(torch.zeros(1, in_features, out_features, requires_grad=True))
        self.alpha = nn.Parameter(torch.zeros(self.k, in_features, out_features, requires_grad=True))
        self.tau = nn.Parameter(torch.full((self.k, in_features, out_features), -5.0, requires_grad=True))
        if intercept:
            self.intercept = nn.Parameter(torch.zeros(out_features, requires_grad=True))
        else:
            self.register_buffer('intercept', None)

        if negative_relationships:
            if shared_directions == 0:
                raise ValueError("shared_directions must be greater than 0.")
            if out_features % shared_directions != 0:
                raise ValueError("out_features must be divisible by shared_directions.")
            self.directions = nn.Parameter(torch.zeros(in_features, int(out_features / shared_directions), requires_grad=True))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        sp_tau = F.softplus(self.tau)
        
        a = F.softplus(self.omega)
        for i in range(self.k):
            matrix = torch.zeros((2*(i+1)+1, 2*(i+1)-1, self.input_dim, self.output_dim), device=x.device)
            range_indices = torch.arange(2*(i+1)-1, device=x.device)
            matrix[range_indices, range_indices, :, :] = 1
            matrix[range_indices + 1, range_indices, :, :] = -2 * self.alpha[i]
            matrix[range_indices + 2, range_indices, :, :] = self.alpha[i] ** 2 + sp_tau[i]
            a = torch.einsum('abio,bio->aio', matrix, a)
        
        if self.negative_relationships:
            a.multiply_(self.directions.repeat_interleave(self.shared_directions, dim=1))

        # remove relationship between some items and latent variables
        if self.relationship_matrix is not None:
            a[:, ~self.relationship_matrix] = 0.0
        # a dimensions: (degree, input_dim, output_dim)
        # divide by 1, 2, ..., 2k+1 to get the polynomial coefficients
        b = a / (torch.arange(1, 2*self.k+2, device=x.device).unsqueeze(1).unsqueeze(2))
        x_powers = x.unsqueeze(2) ** torch.arange(1, 2*self.k+2, device=x.device)
        # x_powers dimensions: (batch, input_dim, degree)
        result = torch.einsum('abc,cbd->ad', x_powers, b)
        if self.intercept is not None:
            result += self.intercept
        return result
    
    @torch.no_grad()
    def get_polynomial_coefficients(self) -> torch.Tensor:
        """
        Returns the polynomial coefficients.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor]
            Tuple of tensors containing the coefficients of the polynomial with dimensions (degree, input_dim, output_dim), the second tensor contains the intercept if it exists and None otherwise.
        """
        sp_tau = F.softplus(self.tau)
        
        b = F.softplus(self.omega)
        for i in range(self.k):
            matrix = torch.zeros((2*(i+1)+1, 2*(i+1)-1, self.input_dim, self.output_dim))
            range_indices = torch.arange(2*(i+1)-1)
            matrix[range_indices, range_indices, :, :] = 1
            matrix[range_indices + 1, range_indices, :, :] = -2 * self.alpha[i]
            matrix[range_indices + 2, range_indices, :, :] = self.alpha[i] ** 2 + sp_tau[i]
            b = torch.einsum('abio,bio->aio', matrix, b) / (i + 1)

        if self.negative_relationships:
            b.multiply_(self.directions.repeat_interleave(self.shared_directions, dim=1))

        # remove relationship between some items and latent variables
        if self.relationship_matrix is not None:
            b[:, ~self.relationship_matrix] = 0.0

        return b, self.intercept
